# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

#
# Basic functionality for regression tests
#

__all__ = [
    'CompileOnlyRegressionTest', 'RegressionTest', 'RunOnlyRegressionTest',
    'RegressionMixin'
]


import glob
import hashlib
import inspect
import itertools
import numbers
import os
import shutil

import reframe.core.fields as fields
import reframe.core.hooks as hooks
import reframe.core.logging as logging
import reframe.core.runtime as rt
import reframe.utility as util
import reframe.utility.jsonext as jsonext
import reframe.utility.osext as osext
import reframe.utility.sanity as sn
import reframe.utility.typecheck as typ
import reframe.utility.udeps as udeps
from reframe.core.backends import getlauncher, getscheduler
from reframe.core.buildsystems import BuildSystemField
from reframe.core.containers import (ContainerPlatform, ContainerPlatformField)
from reframe.core.deferrable import (_DeferredExpression,
                                     _DeferredPerformanceExpression)
from reframe.core.environments import Environment
from reframe.core.exceptions import (BuildError, DependencyError,
                                     PerformanceError, PipelineError,
                                     SanityError, SkipTestError,
                                     ReframeSyntaxError)
from reframe.core.meta import RegressionTestMeta
from reframe.core.schedulers import Job


class _NoRuntime(ContainerPlatform):
    '''Proxy container runtime for storing container platform info early enough.

    This will be replaced by the framework with a concrete implementation
    based on the current partition info.
    '''

    def emit_prepare_commands(self, stagedir):
        raise NotImplementedError

    def launch_command(self, stagedir):
        raise NotImplementedError


# Meta-type to use for type checking in some variables
Deferrable = typ.make_meta_type('Deferrable', _DeferredExpression)


# Valid systems/environments mini-language
_N = r'(\w[-.\w]*)'         # name
_NW = rf'(\*|{_N})'         # name or wildcard
_F = rf'([+-]{_N})'         # feature
_OP = r'([=<>]|!=|>=|<=)'   # relational operator (unused for the moment)
_KV = rf'(%{_N}=\S+)'       # key/value pair
_FKV = rf'({_F}|{_KV})'     # feature | key/value pair
_VALID_ENV_SYNTAX = rf'^({_NW}|{_FKV}(\s+{_FKV})*)$'

_S = rf'({_NW}(:{_NW})?)'   # system/partition
_VALID_SYS_SYNTAX = rf'^({_S}|{_FKV}(\s+{_FKV})*)$'


_PIPELINE_STAGES = (
    '__init__',
    'setup',
    'compile', 'compile_wait',
    'run', 'run_wait',
    'sanity',
    'performance',
    'cleanup'
)


_USER_PIPELINE_STAGES = (
    'init', 'setup', 'compile', 'run', 'sanity', 'performance', 'cleanup'
)


_RFM_TEST_KIND_MIXIN = 0
_RFM_TEST_KIND_COMPILE = 1
_RFM_TEST_KIND_RUN = 2


class RegressionMixin(metaclass=RegressionTestMeta):
    '''Base mixin class for regression tests.

    Multiple inheritance from more than one
    :class:`RegressionTest` class is not allowed in ReFrame. Hence, mixin
    classes provide the flexibility to bundle reusable test add-ons, leveraging
    the metaclass magic implemented in
    :class:`RegressionTestMeta`. Using this metaclass allows mixin classes to
    use powerful ReFrame features, such as hooks, parameters or variables.

    .. versionadded:: 3.4.2
    '''

    _rfm_regression_class_kind = _RFM_TEST_KIND_MIXIN


class RegressionTest(RegressionMixin, jsonext.JSONSerializable):
    '''Base class for regression tests.

    All regression tests must eventually inherit from this class.
    This class provides the implementation of the pipeline phases that the
    regression test goes through during its lifetime.

    This class accepts parameters at the *class definition*, i.e., the test
    class can be defined as follows:

    .. code-block:: python

       class MyTest(RegressionTest, param='foo', ...):

    where ``param`` is one of the following:

    :param pin_prefix: lock the test prefix to the directory where the current
        class lives.

    :param require_version: a list of ReFrame version specifications that this
        test is allowed to run. A version specification string can have one of
        the following formats:

        - ``VERSION``: Specifies a single version.
        - ``{OP}VERSION``, where ``{OP}`` can be any of ``>``, ``>=``, ``<``,
          ``<=``, ``==`` and ``!=``. For example, the version specification
          string ``'>=3.5.0'`` will allow the following test to be loaded
          only by ReFrame 3.5.0 and higher. The ``==VERSION`` specification
          is the equivalent of ``VERSION``.
        - ``V1..V2``: Specifies a range of versions.

        The test will be selected if *any* of the versions is satisfied, even
        if the versions specifications are conflicting.

    :param special: allow pipeline stage methods to be overriden in this class.

    .. note::
        .. versionchanged:: 2.19
           Base constructor takes no arguments.

        .. versionadded:: 3.3
           The ``pin_prefix`` class definition parameter is added.

        .. versionadded:: 3.7.0
           The ``require_verion`` class definition parameter is added.

    .. warning::
        .. versionchanged:: 3.4.2
           Multiple inheritance with a shared common ancestor is not allowed.

    '''

    _rfm_regression_class_kind = _RFM_TEST_KIND_COMPILE | _RFM_TEST_KIND_RUN

    def disable_hook(self, hook_name):
        '''Disable pipeline hook by name.

        :arg hook_name: The function name of the hook to be disabled.

        :meta private:
        '''
        self._disabled_hooks.add(hook_name)

    @classmethod
    def pipeline_hooks(cls):
        ret = {}
        for hook in cls._rfm_hook_registry:
            for stage in hook.stages:
                try:
                    ret[stage].append(hook.fn)
                except KeyError:
                    ret[stage] = [hook.fn]

        return ret

    #: List of programming environments supported by this test.
    #:
    #: The syntax of this attribute is exactly the same as of the
    #: :attr:`valid_systems` except that the ``a:b`` entries are invalid.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``required``
    #:
    #: .. seealso::
    #:    - `Environment features
    #:      <config_reference.html#environments-.features>`__
    #:    - `Environment extras
    #:      <config_reference.html#environments-.extras>`__
    #:
    #: .. versionchanged:: 2.12
    #:    Programming environments can now be specified using wildcards.
    #:
    #: .. versionchanged:: 2.17
    #:    Support for wildcards is dropped.
    #:
    #: .. versionchanged:: 3.3
    #:    Default value changed from ``[]`` to ``None``.
    #:
    #: .. versionchanged:: 3.6
    #:    Default value changed from ``None`` to ``required``.
    #:
    #: .. versionchanged:: 3.11.0
    #:    Extend syntax to support features and key/value pairs.
    valid_prog_environs = variable(typ.List[typ.Str[_VALID_ENV_SYNTAX]],
                                   loggable=True)

    #: List of systems or system features or system properties required by this
    #: test.
    #:
    #: Each entry in this list is a requirement and can have one of the
    #: following forms:
    #:
    #: - ``sysname``: The test is valid for system named ``sysname``.
    #: - ``sysname:partname``: The test is valid for the partition ``partname``
    #:   of system ``sysname``.
    #: - ``*``: The test is valid for any system.
    #: - ``*:partname``: The test is valid for any partition named ``partname``
    #:   in any system.
    #: - ``+feat``: The test is valid for all partitions that define feature
    #:   ``feat`` as a feature.
    #: - ``-feat``: The test is valid for all partitions that do not define
    #:   feature ``feat`` as a feature.
    #: - ``%key=val``: The test is valid for all partitions that define the
    #:   extra property ``key`` with the value ``val``.
    #:
    #: Multiple features and key/value pairs can be included in a single entry
    #: of the :attr:`valid_systems` list, in which case an AND operation on
    #: these constraints is implied. For example, the test defining the
    #: following will be valid for all systems that have define both ``feat1``
    #: and ``feat2`` and set ``foo=1``
    #:
    #: .. code-block:: python
    #:
    #:    valid_systems = ['+feat1 +feat2 %foo=1']
    #:
    #: For key/value pairs comparisons, ReFrame will automatically convert the
    #: value in the key/value spec to the type of the value of the
    #: corresponding entry in the partitions ``extras`` property. In the above
    #: example, if the type of ``foo`` property is integer, ``1`` will be
    #: converted to an integer value. If a conversion to the target type is not
    #: possible, then the requested key/value pair is not matched.
    #:
    #: Multiple entries in the :attr:`valid_systems` list are implicitly ORed,
    #: such that the following example implies that the test is valid for
    #: either ``sys1`` or for any other system that does not define ``feat``.
    #:
    #: .. code-block:: python
    #:
    #:    valid_systems = ['sys1', '-feat']
    #:
    #: :type: :class:`List[str]`
    #: :default: ``None``
    #:
    #: .. seealso::
    #:    - `System partition features
    #:      <config_reference.html#systems-.partitions-.features>`__
    #:    - `System partition extras
    #:      <config_reference.html#systems-.partitions-.extras>`__
    #:
    #: .. versionchanged:: 3.3
    #:    Default value changed from ``[]`` to ``None``.
    #:
    #: .. versionchanged:: 3.6
    #:    Default value changed from ``None`` to ``required``.
    #:
    #:  .. versionchanged:: 3.11.0
    #:     Extend syntax to support features and key/value pairs.
    valid_systems = variable(typ.List[typ.Str[_VALID_SYS_SYNTAX]],
                             loggable=True)

    #: A detailed description of the test.
    #:
    #: :type: :class:`str`
    #: :default: ``''``
    #:
    #: .. versionchanged:: 4.0
    #:    The default value is now the empty string.
    descr = variable(str, value='', loggable=True)

    #: The path to the source file or source directory of the test.
    #:
    #: It must be a path relative to the :attr:`sourcesdir`, pointing to a
    #: subfolder or a file contained in :attr:`sourcesdir`. This applies also
    #: in the case where :attr:`sourcesdir` is a Git repository.
    #:
    #: If it refers to a regular file, this file will be compiled using the
    #: :class:`SingleSource <reframe.core.buildsystems.SingleSource>` build
    #: system.
    #: If it refers to a directory, ReFrame will try to infer the build system
    #: to use for the project and will fall back in using the :class:`Make
    #: <reframe.core.buildsystems.Make>` build system, if it cannot find a more
    #: specific one.
    #:
    #: :type: :class:`str`
    #: :default: ``''``
    sourcepath = variable(str, value='', loggable=True)

    #: The directory containing the test's resources.
    #:
    #: This directory may be specified with an absolute path or with a path
    #: relative to the location of the test. Its contents will always be copied
    #: to the stage directory of the test.
    #:
    #: This attribute may also accept a URL, in which case ReFrame will treat
    #: it as a Git repository and will try to clone its contents in the stage
    #: directory of the test.
    #:
    #: If set to :class:`None`, the test has no resources an no action is
    #: taken.
    #:
    #: :type: :class:`str` or :class:`None`
    #: :default: ``'src'`` if such a directory exists at the test level,
    #:    otherwise ``None``
    #:
    #: .. note::
    #:     .. versionchanged:: 2.9
    #:        Allow :class:`None` values to be set also in regression tests
    #:        with a compilation phase
    #:
    #:     .. versionchanged:: 2.10
    #:        Support for Git repositories was added.
    #:
    #:     .. versionchanged:: 3.0
    #:        Default value is now conditionally set to either ``'src'`` or
    #:        :class:`None`.
    sourcesdir = variable(str, type(None), value='src', loggable=True)

    #: .. versionadded:: 2.14
    #:
    #: The build system to be used for this test.
    #: If not specified, the framework will try to figure it out automatically
    #: based on the value of :attr:`sourcepath`.
    #:
    #: This field may be set using either a string referring to a concrete
    #: build system class name
    #: (see `build systems <#build-systems>`__) or an instance of
    #: :class:`reframe.core.buildsystems.BuildSystem`. The former is the
    #: recommended way.
    #:
    #:
    #: :type: :class:`str` or :class:`reframe.core.buildsystems.BuildSystem`.
    #: :default: :class:`None`.
    build_system = variable(type(None), field=BuildSystemField, value=None)

    #: .. versionadded:: 3.0
    #:
    #: List of shell commands to be executed before compiling.
    #:
    #: These commands are emitted in the build script before the actual build
    #: commands generated by the selected `build system
    #: <#reframe.core.pipeline.RegressionTest.build_system>`__.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    prebuild_cmds = variable(typ.List[str], value=[], loggable=True)

    #: .. versionadded:: 3.0
    #:
    #: List of shell commands to be executed after a successful compilation.
    #:
    #: These commands are emitted in the script after the actual build
    #: commands generated by the selected `build system
    #: <#reframe.core.pipeline.RegressionTest.build_system>`__.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    postbuild_cmds = variable(typ.List[str], value=[], loggable=True)

    #: The name of the executable to be launched during the run phase.
    #:
    #: If this variable is undefined when entering the compile pipeline
    #: stage, it will be set to ``os.path.join('.', self.unique_name)``.
    #: Classes that override the compile stage may leave this variable
    #: undefined.
    #:
    #: :type: :class:`str`
    #: :default: :class:`required`
    #:
    #: .. versionchanged:: 3.7.3
    #:    Default value changed from ``os.path.join('.', self.unique_name)`` to
    #:    :class:`required`.
    executable = variable(str, loggable=True)

    #: List of options to be passed to the :attr:`executable`.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    executable_opts = variable(typ.List[str], value=[], loggable=True)

    #: .. versionadded:: 2.20
    #:
    #: The container platform to be used for launching this test.
    #:
    #: This field is set automatically by the default container runtime
    #: associated with the current system partition. Users may also set this,
    #: explicitly overriding any partition setting. If the
    #: :attr:`~reframe.core.containers.ContainerPlatform.image` attribute of
    #: :attr:`container_platform` is set, then the test will run inside a
    #: container using the specified container runtime.
    #:
    #: .. code:: python
    #:
    #:    self.container_platform = 'Singularity'
    #:    self.container_platform.image = 'docker://ubuntu:18.04'
    #:    self.container_platform.command = 'cat /etc/os-release'
    #:
    #: If the test will run inside a container, the :attr:`executable` and
    #: :attr:`executable_opts` attributes are ignored. The container platform's
    #: :attr:`~reframe.core.containers.ContainerPlatform.command` will be used
    #: instead.
    #:
    #: .. note::
    #:
    #:    Only the run phase of the test will run inside the container.
    #:    If you enable the containerized run in a non run-only test, the
    #:    compilation phase will still run natively.
    #:
    #: :type: :class:`str` or
    #:     :class:`~reframe.core.containers.ContainerPlatform`.
    #: :default: the container runtime specified in the current system
    #:   partition's configuration (see also
    #:   :ref:`container-platform-configuration`).
    #:
    #: .. versionchanged:: 3.12.0
    #:    This field is now set automatically from the current partition's
    #:    configuration.
    container_platform = variable(field=ContainerPlatformField,
                                  value=_NoRuntime())

    #: .. versionadded:: 3.0
    #:
    #: List of shell commands to execute before the parallel launch command.
    #:
    #: These commands do not execute in the context of ReFrame.
    #: Instead, they are emitted in the generated job script just before the
    #: actual job launch command.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    prerun_cmds = variable(typ.List[str], value=[], loggable=True)

    #: .. versionadded:: 3.0
    #:
    #: List of shell commands to execute after the parallel launch command.
    #:
    #: See :attr:`prerun_cmds` for a more detailed description of the
    #: semantics.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    postrun_cmds = variable(typ.List[str], value=[], loggable=True)

    #: List of files to be kept after the test finishes.
    #:
    #: By default, the framework saves the standard output, the standard error
    #: and the generated shell script that was used to run this test.
    #:
    #: These files will be copied over to the test's output directory
    #: during the :func:`cleanup` phase.
    #:
    #: Directories are also accepted in this field.
    #:
    #: Relative path names are resolved against the stage directory.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    #:
    #: .. versionchanged:: 3.3
    #:    This field accepts now also file glob patterns.
    #:
    keep_files = variable(typ.List[str], value=[], loggable=True)

    #: List of files or directories (relative to the :attr:`sourcesdir`) that
    #: will be symlinked in the stage directory and not copied.
    #:
    #: You can use this variable to avoid copying very large files to the stage
    #: directory.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    readonly_files = variable(typ.List[str], value=[], loggable=True)

    #: Set of tags associated with this test.
    #:
    #: This test can be selected from the frontend using any of these tags.
    #:
    #: :type: :class:`Set[str]`
    #: :default: an empty set
    tags = variable(typ.Set[str], value=set(), loggable=True)

    #: List of people responsible for this test.
    #:
    #: When the test fails, this contact list will be printed out.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    maintainers = variable(typ.List[str], value=[], loggable=True)

    #: Mark this test as a strict performance test.
    #:
    #: If a test is marked as non-strict, the performance checking phase will
    #: always succeed, unless the ``--strict`` command-line option is passed
    #: when invoking ReFrame.
    #:
    #: :type: boolean
    #: :default: :class:`True`
    strict_check = variable(typ.Bool, value=True, loggable=True)

    #: Number of tasks required by this test.
    #:
    #: If the number of tasks is set to a number ``<=0``, ReFrame will try to
    #: flexibly allocate the number of tasks, based on the command line option
    #: |--flex-alloc-nodes|_. A negative number is used to indicate the minimum
    #: number of tasks required for the test. In this case the minimum number
    #: of tasks is the absolute value of the number, while Setting
    #: :attr:`num_tasks` to ``0`` is equivalent to setting it to
    #: :attr:`-num_tasks_per_node
    #: <reframe.core.pipeline.RegressionTest.num_tasks_per_node>`.
    #:
    #: :type: integral
    #: :default: ``1``
    #:
    #: .. note::
    #:     .. versionchanged:: 2.15
    #:        Added support for flexible allocation of the number of tasks
    #:        if the number of tasks is set to ``0``.
    #:     .. versionchanged:: 2.16
    #:        Negative :attr:`num_tasks` is allowed for specifying the minimum
    #:        number of required tasks by the test.
    #:
    #: .. |--flex-alloc-nodes| replace:: :attr:`--flex-alloc-nodes`
    #: .. _--flex-alloc-nodes: manpage.html#cmdoption-flex-alloc-nodes
    num_tasks = variable(int, value=1, loggable=True)

    #: Number of tasks per node required by this test.
    #:
    #: Ignored if :class:`None`.
    #:
    #: :type: integral or :class:`None`
    #: :default: :class:`None`
    num_tasks_per_node = variable(int, type(None), value=None, loggable=True)

    #: Number of GPUs per node required by this test.
    #: This attribute is translated internally to the ``_rfm_gpu`` resource.
    #: For more information on test resources, have a look at the
    #: :attr:`extra_resources` attribute.
    #:
    #: :type: integral or :const:`None`
    #: :default: :const:`None`
    #:
    #: .. versionchanged:: 4.0.0
    #:    The default value changed to :const:`None`.
    num_gpus_per_node = variable(int, type(None), value=None, loggable=True)

    #: Number of CPUs per task required by this test.
    #:
    #: Ignored if :class:`None`.
    #:
    #: :type: integral or :class:`None`
    #: :default: :class:`None`
    num_cpus_per_task = variable(int, type(None), value=None, loggable=True)

    #: Number of tasks per core required by this test.
    #:
    #: Ignored if :class:`None`.
    #:
    #: :type: integral or :class:`None`
    #: :default: :class:`None`
    num_tasks_per_core = variable(int, type(None), value=None, loggable=True)

    #: Number of tasks per socket required by this test.
    #:
    #: Ignored if :class:`None`.
    #:
    #: :type: integral or :class:`None`
    #: :default: :class:`None`
    num_tasks_per_socket = variable(int, type(None), value=None, loggable=True)

    #: Specify whether this tests needs simultaneous multithreading enabled.
    #:
    #: Ignored if :class:`None`.
    #:
    #: :type: boolean or :class:`None`
    #: :default: :class:`None`
    use_multithreading = variable(
        typ.Bool, type(None), value=None, loggable=True)

    #: .. versionadded:: 3.0
    #:
    #: The maximum time a job can be pending before starting running.
    #:
    #: Time duration is specified as of the :attr:`time_limit` attribute.
    #:
    #: :type: :class:`str` or :class:`datetime.timedelta`
    #: :default: :class:`None`
    max_pending_time = variable(
        type(None), field=fields.TimerField, value=None, loggable=True
    )

    #: Specify whether this test needs exclusive access to nodes.
    #:
    #: :type: boolean
    #: :default: :class:`False`
    exclusive_access = variable(typ.Bool, value=False, loggable=True)

    #: Always execute this test locally.
    #:
    #: :type: boolean
    #: :default: :class:`False`
    local = variable(typ.Bool, value=False, loggable=True)

    #: The set of reference values for this test.
    #:
    #: The reference values are specified as a scoped dictionary keyed on the
    #: performance variables defined in :attr:`perf_patterns` and scoped under
    #: the system/partition combinations.
    #: The reference itself is a four-tuple that contains the reference value,
    #: the lower and upper thresholds and the measurement unit.
    #:
    #: An example follows:
    #:
    #: .. code:: python
    #:
    #:    self.reference = {
    #:        'sys0:part0': {
    #:            'perfvar0': (50, -0.1, 0.1, 'Gflop/s'),
    #:            'perfvar1': (20, -0.1, 0.1, 'GB/s')
    #:        },
    #:        'sys0:part1': {
    #:            'perfvar0': (100, -0.1, 0.1, 'Gflop/s'),
    #:            'perfvar1': (40, -0.1, 0.1, 'GB/s')
    #:        }
    #:    }
    #:
    #: To better understand how to set the performance reference tuple, here
    #: are some examples with both positive and negative reference values:
    #:
    #:   ============================== ============  ==========  ===========
    #:   **Performance Tuple**          **Expected**  **Lowest**  **Highest**
    #:   ``(100, -0.01, 0.02, 'MB/s')`` 100 MB/s      99 MB/s     102 MB/s
    #:   ``(100, -0.01, None, 'MB/s')`` 100 MB/s      99 MB/s     inf MB/s
    #:   ``(100, None, 0.02, 'MB/s')``  100 MB/s      -inf MB/s   102 MB/s
    #:   ``(-100, -0.01, 0.02, 'C')``     -100 C        -101 C      -98 C
    #:   ``(-100, -0.01, None, 'C')``     -100 C        -101 C      inf C
    #:   ``(-100, None, 0.02, 'C')``      -100 C        -inf C      -98 C
    #:   ============================== ============  ==========  ===========
    #:
    #: During the performance stage of the pipeline, the reference tuple
    #: elements, except the unit, are passed to the
    #: :func:`~reframe.utility.sanity.assert_reference` function along with the
    #: obtained performance value in order to actually assess whether the test
    #: passes the performance check or not.
    #:
    #: :type: A scoped dictionary with system names as scopes, performance
    #:   variables as keys and reference tuples as values.
    #:   The elements of reference tuples cannot be deferrable expressions.
    #:
    #: .. note::
    #:     .. versionchanged:: 3.0
    #:        The measurement unit is required. The user should explicitly
    #:        specify :class:`None` if no unit is available.
    #:
    #:     .. versionchanged:: 3.8.0
    #:        The unit in the reference tuple is again optional, but it is
    #:        recommended to use it for clarity.
    #:
    #:     .. versionchanged:: 4.0.0
    #:        Deferrable expressions are not allowed in reference tuples.
    reference = variable(
        typ.Tuple[~Deferrable, ~Deferrable, ~Deferrable],
        typ.Dict[str, typ.Dict[str, typ.Tuple[~Deferrable, ~Deferrable,
                                              ~Deferrable, ~Deferrable]]],
        field=fields.ScopedDictField, value={}
    )

    #: Require that a reference is defined for each system that this test is
    #: run on.
    #:
    #: If this is set and a reference is not found for the current system, the
    #: test will fail.
    #:
    #: :type: boolean
    #: :default: :const:`False`
    #:
    #: .. versionadded:: 4.0.0
    require_reference = variable(typ.Bool, value=False)

    #:
    #: Refer to the :doc:`ReFrame Tutorials </tutorials>` for concrete usage
    #: examples.
    #:
    #: If not set, a sanity error may be raised during sanity checking if no
    #: other sanity checking functions already exist.
    #:
    #: :type: A deferrable expression (i.e., the result of a :doc:`sanity
    #:     function </deferrable_functions_reference>`)
    #: :default: :class:`required`
    #:
    #: .. note::
    #:    .. versionchanged:: 2.9
    #:       The default behaviour has changed and it is now considered a
    #:       sanity failure if this attribute is set to :class:`required`.
    #:
    #:       If a test doesn't care about its output, this must be stated
    #:       explicitly as follows:
    #:
    #:       ::
    #:
    #:           self.sanity_patterns = sn.assert_true(1)
    #:
    #:    .. versionchanged:: 3.6
    #:       The default value has changed from ``None`` to ``required``.
    sanity_patterns = variable(_DeferredExpression)

    #: Patterns for verifying the performance of this test.
    #:
    #: If set to :class:`None`, no performance checking will be performed.
    #:
    #: :type: A dictionary with keys of type :class:`str` and deferrable
    #:     expressions (i.e., the result of a :doc:`sanity function
    #:     </deferrable_functions_reference>`) as values.
    #:     :class:`None` is also allowed.
    #: :default: :class:`None`
    #:
    #: .. warning::
    #:
    #:    You are advised to follow the new syntax for defining performance
    #:    variables in your tests using either the :func:`@performance_function
    #:    <reframe.core.builtins.performance_function>` builtin or the
    #:    :attr:`perf_variables`, as :attr:`perf_patterns` will likely be
    #:    deprecated in the future.
    perf_patterns = variable(typ.Dict[str, _DeferredExpression], type(None))

    #: The performance variables associated with the test.
    #:
    #: In this context, a performance variable is a key-value pair, where the
    #: key is the desired variable name and the value is the deferred
    #: performance expression (i.e. the result of a :ref:`deferrable
    #: performance function<deferrable-performance-functions>`) that computes
    #: or extracts the performance variable's value.
    #:
    #: By default, ReFrame will populate this field during the test's
    #: instantiation with all the member functions decorated with the
    #: :func:`@performance_function
    #: <reframe.core.builtins.performance_function>` decorator.
    #: If no performance functions are present in the class, no performance
    #: checking or reporting will be carried out.
    #:
    #: This mapping may be extended or replaced by other performance variables
    #: that may be defined in any pipeline hook executing before the
    #: performance stage. To this end, deferred performance functions can be
    #: created inline using the utility
    #: :func:`~reframe.utility.sanity.make_performance_function`.
    #:
    #: Refer to the :doc:`ReFrame Tutorials </tutorials>` for concrete usage
    #: examples.
    #:
    #: :type: A dictionary with keys of type :class:`str` and deferred
    #:     performance expressions as values (see
    #:     :ref:`deferrable-performance-functions`).
    #: :default: Collection of performance variables associated to each of
    #:     the member functions decorated with the :func:`@performance_function
    #:     <reframe.core.builtins.performance_function>`
    #:     decorator.
    #:
    #: .. versionadded:: 3.8.0
    perf_variables = variable(typ.Dict[str, _DeferredPerformanceExpression],
                              value={})

    #: List of modules to be loaded before running this test.
    #:
    #: These modules will be loaded during the :func:`setup` phase.
    #:
    #: :type: :class:`List[str]` or :class:`Dict[str, object]`
    #: :default: ``[]``
    modules = variable(typ.List[str], typ.List[typ.Dict[str, object]],
                       value=[], loggable=True)

    #: Environment variables to be set before running this test.
    #:
    #: The value of the environment variables can be of any type. ReFrame will
    #: invoke :func:`str` on it whenever it needs to emit it in a script.
    #:
    #: :type: :class:`Dict[str, object]`
    #: :default: ``{}``
    #:
    #: .. versionadded:: 4.0.0
    env_vars = variable(typ.Dict[str, str],
                        typ.Dict[str, object], value={}, loggable=True)
    # NOTE: We still keep the original type, just to allow setting this
    # variable from the command line, because otherwise, ReFrame will not know
    # how to convert a value to an arbitrary object.

    #: Environment variables to be set before running this test.
    #:
    #: This is an alias of :attr:`env_vars`.
    #:
    #: .. deprecated:: 4.0.0
    #:    Please use :attr:`env_vars` instead.
    variables = deprecate(variable(alias=env_vars, loggable=True),
                          f"the use of 'variables' is deprecated; "
                          f"please use 'env_vars' instead")

    #: Time limit for this test.
    #:
    #: Time limit is specified as a string in the form
    #: ``<days>d<hours>h<minutes>m<seconds>s`` or as number of seconds. If set
    #: to :class:`None`, the
    #: :attr:`~reframe.core.systems.SystemPartition.time_limit` of the current
    #: system partition will be used.
    #:
    #: :type: :class:`str` or :class:`float` or :class:`int`
    #: :default: :class:`None`
    #:
    #: .. note::
    #:    .. versionchanged:: 2.15
    #:       This attribute may be set to :class:`None`.
    #:
    #: .. warning::
    #:    .. versionchanged:: 3.0
    #:       The old syntax using a ``(h, m, s)`` tuple is deprecated.
    #:
    #:    .. versionchanged:: 3.2
    #:       - The old syntax using a ``(h, m, s)`` tuple is dropped.
    #:       - Support of `timedelta` objects is dropped.
    #:       - Number values are now accepted.
    #:
    #:    .. versionchanged:: 3.5.1
    #:       The default value is now :class:`None` and it can be set globally
    #:       per partition via the configuration.
    time_limit = variable(type(None), field=fields.TimerField,
                          value=None, loggable=True)

    #: .. versionadded:: 3.5.1
    #:
    #: The time limit for the build job of the regression test.
    #:
    #: It is specified similarly to the :attr:`time_limit` attribute.
    #:
    #: :type: :class:`str` or :class:`float` or :class:`int`
    #: :default: :class:`None`
    build_time_limit = variable(type(None), field=fields.TimerField,
                                value=None, loggable=True)

    #: .. versionadded:: 2.8
    #:
    #: Extra resources for this test.
    #:
    #: This field is for specifying custom resources needed by this test. These
    #: resources are defined in the `configuration
    #: <config_reference.html#.systems[].partitions[].resources>`__ of a system
    #: partition. For example, assume that two additional resources, named
    #: ``gpu`` and ``datawarp``, are defined in the configuration file as
    #: follows:
    #:
    #: ::
    #:
    #:     'resources': [
    #:         {
    #:             'name': 'gpu',
    #:             'options': ['--gres=gpu:{num_gpus_per_node}']
    #:         },
    #:         {
    #:             'name': 'datawarp',
    #:             'options': [
    #:                 '#DW jobdw capacity={capacity}',
    #:                 '#DW stage_in source={stagein_src}'
    #:             ]
    #:         }
    #:     ]
    #:
    #: A regression test may then instantiate the above resources by setting
    #: the :attr:`extra_resources` attribute as follows:
    #:
    #: ::
    #:
    #:     self.extra_resources = {
    #:         'gpu': {'num_gpus_per_node': 2}
    #:         'datawarp': {
    #:             'capacity': '100GB',
    #:             'stagein_src': '/foo'
    #:         }
    #:     }
    #:
    #: The generated batch script (for Slurm) will then contain the following
    #: lines:
    #:
    #: ::
    #:
    #:     #SBATCH --gres=gpu:2
    #:     #DW jobdw capacity=100GB
    #:     #DW stage_in source=/foo
    #:
    #: Notice that if the resource specified in the configuration uses an
    #: alternative directive prefix (in this case ``#DW``), this will replace
    #: the standard prefix of the backend scheduler (in this case ``#SBATCH``)
    #:
    #: If the resource name specified in this variable does not match a
    #: resource name in the partition configuration, it will be simply ignored.
    #: The :attr:`num_gpus_per_node` attribute translates internally to the
    #: ``_rfm_gpu`` resource, so that setting
    #: ``self.num_gpus_per_node = 2`` is equivalent to the following:
    #:
    #: ::
    #:
    #:     self.extra_resources = {'_rfm_gpu': {'num_gpus_per_node': 2}}
    #:
    #: :type: :class:`Dict[str, Dict[str, object]]`
    #: :default: ``{}``
    #:
    #: .. note::
    #:    .. versionchanged:: 2.9
    #:       A new more powerful syntax was introduced
    #:       that allows also custom job script directive prefixes.
    extra_resources = variable(typ.Dict[str, typ.Dict[str, object]],
                               value={}, loggable=True)

    #: .. versionadded:: 3.3
    #:
    #: Always build the source code for this test locally. If set to
    #: :class:`False`, ReFrame will spawn a build job on the partition where
    #: the test will run. Setting this to :class:`False` is useful when
    #: cross-compilation is not supported on the system where ReFrame is run.
    #: Normally, ReFrame will mark the test as a failure if the spawned job
    #: exits with a non-zero exit code. However, certain scheduler backends,
    #: such as the ``squeue`` do not set it. In such cases, it is the user's
    #: responsibility to check whether the build phase failed by adding an
    #: appropriate sanity check.
    #:
    #: :type: boolean
    #: :default: :class:`True`
    build_locally = variable(typ.Bool, value=True, loggable=True)

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)

        # Determine the prefix
        try:
            prefix = cls._rfm_custom_prefix
        except AttributeError:
            if osext.is_interactive():
                prefix = os.getcwd()
            else:
                try:
                    prefix = cls._rfm_pinned_prefix
                except AttributeError:
                    prefix = os.path.abspath(
                        os.path.dirname(inspect.getfile(cls))
                    )

        # Prepare initialization of test defaults (variables and parameters are
        # injected after __new__ has returned, so we schedule this function
        # call as a pre-init hook).
        obj.__deferred_rfm_init = obj.__rfm_init__(prefix)

        # Build pipeline hook registry and add the pre-init hook
        cls._rfm_pipeline_hooks = cls._process_hook_registry()
        cls._rfm_pipeline_hooks['pre___init__'] = [obj.__pre_init__]

        # Attach the hooks to the pipeline stages
        for stage in _PIPELINE_STAGES:
            cls._add_hooks(stage)

        return obj

    @final
    def __pre_init__(self):
        '''Initialize the test defaults from a pre-init hook.'''
        self.__deferred_rfm_init.evaluate()

        # Build the default performance dict
        if not self.perf_variables:
            for fn in self._rfm_perf_fns.values():
                self.perf_variables[fn._rfm_perf_key] = fn(self)

    def __init__(self):
        pass

    @classmethod
    def __init_subclass__(cls, *, special=False, pin_prefix=False,
                          require_version=None, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._rfm_override_final = special

        if require_version:
            cls._rfm_required_version = require_version
        elif not hasattr(cls, '_rfm_required_version'):
            cls._rfm_required_version = []

        # Insert the prefix to pin the test to if the test lives in a test
        # library with resources in it.
        if pin_prefix:
            cls._rfm_pinned_prefix = os.path.abspath(
                os.path.dirname(inspect.getfile(cls))
            )

    @deferrable
    def __rfm_init__(self, prefix=None):
        self._perfvalues = {}

        # Static directories of the regression check
        self._prefix = os.path.abspath(prefix)
        if (self.sourcesdir == 'src' and
            not os.path.isdir(os.path.join(self._prefix, self.sourcesdir))):
            self.sourcesdir = None

        # Runtime information of the test
        self._current_partition = None
        self._current_environ = None

        # Associated job
        self._job = None

        # Dynamic paths of the regression check; will be set in setup()
        self._stagedir = None
        self._outputdir = None
        self._stdout = None
        self._stderr = None

        # Compilation process output
        self._build_job = None
        self._compile_proc = None

        # List of dependencies specified by the user
        self._userdeps = []

        # Weak reference to the test case associated with this check
        self._case = None

        if rt.runtime().get_option('general/0/non_default_craype'):
            self._cdt_environ = Environment(
                name='__rfm_cdt_environ',
                env_vars={
                    'LD_LIBRARY_PATH': '$CRAY_LD_LIBRARY_PATH:$LD_LIBRARY_PATH'
                }
            )
        else:
            # Just an empty environment
            self._cdt_environ = Environment('__rfm_cdt_environ')

        # Disabled hooks
        self._disabled_hooks = set()

    @classmethod
    def _process_hook_registry(cls):
        '''Process and validate the pipeline hooks.'''

        _pipeline_hooks = {}
        for stage, hks in cls.pipeline_hooks().items():
            # Pop the stage pre_/post_ prefix
            stage_name = stage.split('_', maxsplit=1)[1]

            if stage_name not in _USER_PIPELINE_STAGES:
                raise ValueError(
                    f'invalid pipeline stage ({stage_name!r}) in class '
                    f'{cls.__qualname__!r}'
                )
            elif stage == 'pre_init':
                raise ValueError(
                    f'{stage} hooks are not allowed ({cls.__qualname__})'
                )
            elif stage == 'post_init':
                stage = 'post___init__'
            elif stage == 'post_compile':
                stage = 'post_compile_wait'
            elif stage == 'post_run':
                stage = 'post_run_wait'

            _pipeline_hooks[stage] = hks

        return _pipeline_hooks

    @classmethod
    def _add_hooks(cls, stage):
        '''Decorate the pipeline stages.'''

        pipeline_hooks = cls._rfm_pipeline_hooks
        fn = getattr(cls, stage)
        new_fn = hooks.attach_hooks(pipeline_hooks)(fn)
        setattr(cls, '_rfm_pipeline_fn_' + stage, new_fn)

    def __getattribute__(self, name):
        if name in _PIPELINE_STAGES:
            name = f'_rfm_pipeline_fn_{name}'

        return super().__getattribute__(name)

    def __getattr__(self, name):
        ''' Intercept the special builtin-related AttributeError.'''

        if (name in self._rfm_var_space and
            not self._rfm_var_space[name].is_defined()):
            raise AttributeError(
                f'required variable {name!r} has not been set'
            ) from None
        else:
            raise AttributeError(
                f'{type(self).__qualname__!r} object has no attribute {name!r}'
            )

    # Export read-only views to interesting fields

    @loggable
    @property
    def unique_name(self):
        '''The unique name of this test.

        :type: :class:`str`

        .. versionadded:: 3.10.0
        '''
        return self._rfm_unique_name

    @loggable
    @property
    def name(self):
        '''The name of the test.

        This is an alias of :attr:`display_name`.
        '''
        return self.display_name

    @loggable
    @property
    def display_name(self):
        '''A human-readable version of the name this test.

        This name contains a string representation of the various parameters
        of this specific test variant.

        :type: :class:`str`

        .. note::
           The display name may not be unique.

        .. versionadded:: 3.10.0

        '''
        def _format_params(cls, info, prefix=' %'):
            name = ''
            for p, v in info['params'].items():
                format_fn = cls.raw_params[p].format
                name += f'{prefix}{p}={format_fn(v)}'

            for f, v in info['fixtures'].items():
                fixt = cls.fixture_space[f]
                if fixt.action == 'join':
                    continue

                name += _format_params(fixt.cls, v, f'{prefix}{f}.')

                # Append any variables set for the fixtures
                for var, val in fixt.variables.items():
                    name += f'{prefix}{f}.{var}={val}'

            return name

        if hasattr(self, '_rfm_display_name'):
            return self._rfm_display_name

        cls = type(self)
        basename = cls.__name__
        variant_info = cls.get_variant_info(self.variant_num, recurse=True)
        self._rfm_display_name = basename + _format_params(cls, variant_info)
        if self.is_fixture():
            # Add the variable info and scope
            fixt_data = self._rfm_fixt_data
            suffix = ''.join(f' %{k}={v}' for k,
                             v in fixt_data.variables.items())
            suffix += f' ~{fixt_data.scope_enc}'
            self._rfm_display_name += suffix

        return self._rfm_display_name

    @loggable
    @property
    def hashcode(self):
        if hasattr(self, '_rfm_hashcode'):
            return self._rfm_hashcode

        m = hashlib.sha256()
        if self.is_fixture():
            m.update(self.unique_name.encode('utf-8'))
        else:
            basename, *params = self.display_name.split(' %')
            m.update(basename.encode('utf-8'))
            for p in sorted(params):
                m.update(p.encode('utf-8'))

        self._rfm_hashcode = m.hexdigest()[:8]
        return self._rfm_hashcode

    @loggable
    @property
    def short_name(self):
        '''A short version of the test's display name.

        The shortened version coincides with the :attr:`unique_name` for
        simple tests and combines the test's class name and a hash code for
        parameterised tests.

        .. versionadded:: 4.0.0

        '''

        if self.unique_name != self.display_name:
            return f'{type(self).__name__}_{self.hashcode}'
        else:
            return self.unique_name

    @property
    def current_environ(self):
        '''The programming environment that the regression test is currently
        executing with.

        This is set by the framework during the :func:`setup` phase.

        :type: :class:`reframe.core.environments.ProgEnvironment`.
        '''
        return self._current_environ

    @property
    def current_partition(self):
        '''The system partition the regression test is currently executing on.

        This is set by the framework during the :func:`setup` phase.

        :type: :class:`reframe.core.systems.SystemPartition`.
        '''
        return self._current_partition

    @property
    def current_system(self):
        '''The system the regression test is currently executing on.

        This is set by the framework during the initialization phase.

        :type: :class:`reframe.core.systems.System`.
        '''
        return rt.runtime().system

    @property
    def variant_num(self):
        '''The variant number of the test.

        This number should be treated as a unique ID representing a unique
        combination of the available parameter and fixture variants.

        :type: :class:`int`
        '''
        return getattr(self, '_rfm_variant_num', None)

    @property
    def param_variant(self):
        '''The point in the parameter space for the test.

        This can be seen as an index to the paraemter space representing a
        unique combination of the parameter values. This number is directly
        mapped from ``variant_num``.

        :type: :class:`int`
        '''
        return getattr(self, '_rfm_param_variant', None)

    @property
    def fixture_variant(self):
        '''The point in the fixture space for the test.

        This can be seen as an index to the fixture space representing a
        unique combination of the fixture variants. This number is directly
        mapped from ``variant_num``.

        :type: :class:`int`
        '''
        return getattr(self, '_rfm_fixt_variant', None)

    def set_var_default(self, name, value):
        '''Set the default value of a variable if variable is undefined.

        A variable is undefined if it is declared and required and no value is
        yet assigned to it.

        :param name: The name of the variable.
        :param value: The value to set the variable to.
        :raises ValueError: If the variable does not exist

        .. versionadded:: 3.10.1

        '''
        var_space = type(self).var_space
        if name not in var_space:
            raise ValueError(f'no such variable: {name!r}')

        if not var_space[name].is_defined():
            setattr(self, name, value)

    @loggable
    @property
    def perfvalues(self):
        return util.MappingView(self._perfvalues)

    @property
    def job(self):
        '''The job descriptor associated with this test.

        This is set by the framework during the :func:`setup` phase.

        :type: :class:`reframe.core.schedulers.Job`.
        '''
        return self._job

    @property
    def logger(self):
        '''A logger associated with this test.

        You can use this logger to log information for your test.
        '''
        return logging.getlogger()

    @loggable
    @property
    def prefix(self):
        '''The prefix directory of the test.

        :type: :class:`str`.
        '''
        return self._prefix

    @loggable
    @property
    def stagedir(self):
        '''The stage directory of the test.

        This is set during the :func:`setup` phase.

        :type: :class:`str`.
        '''
        return self._stagedir

    @loggable
    @property
    def outputdir(self):
        '''The output directory of the test.

        This is set during the :func:`setup` phase.

        .. versionadded:: 2.13

        :type: :class:`str`.
        '''
        return self._outputdir

    @property
    @deferrable
    def stdout(self):
        '''The name of the file containing the standard output of the test.

        This is set during the :func:`setup` phase.

        This attribute is evaluated lazily, so it can by used inside sanity
        expressions.

        :type: :class:`str` or :class:`None` if a run job has not yet been
            created.
        '''
        return self.job.stdout if self.job else None

    @property
    @deferrable
    def stderr(self):
        '''The name of the file containing the standard error of the test.

        This is set during the :func:`setup` phase.

        This attribute is evaluated lazily, so it can by used inside sanity
        expressions.

        :type: :class:`str` or :class:`None` if a run job has not yet been
            created.
        '''
        return self.job.stderr if self.job else None

    @property
    def build_job(self):
        return self._build_job

    @property
    @deferrable
    def build_stdout(self):
        return self.build_job.stdout if self.build_job else None

    @property
    @deferrable
    def build_stderr(self):
        return self.build_job.stderr if self.build_job else None

    # Various properties useful only for logging

    @loggable_as('system')
    @property
    def _system_name(self):
        return self.current_system.name

    @loggable_as('partition')
    @property
    def _partition_name(self):
        if self.current_partition:
            return self.current_partition.name

    @loggable_as('environ')
    @property
    def _environ_name(self):
        if self.current_environ:
            return self.current_environ.name

    @loggable_as('jobid')
    @property
    def _jobid(self):
        if self.job:
            return self.job.jobid

    @loggable_as('job_completion_time_unix')
    @property
    def _job_completion_time(self):
        if self.job:
            return self.job.completion_time

    @loggable_as('job_exitcode')
    @property
    def _job_exitcode(self):
        if self.job:
            return self.job.exitcode

    @loggable_as('job_nodelist')
    @property
    def _job_nodelist(self):
        if self.job:
            return self.job.nodelist

    def info(self):
        '''Provide live information for this test.

        This method is used by the front-end to print the status message
        during the test's execution. This function is also called to provide
        the message for the `check_info
        <config_reference.html#.logging[].handlers[].format>`__ logging
        attribute.
        By default, it returns a message reporting the test name, the current
        partition and the current programming environment that the test is
        currently executing on.

        .. versionadded:: 2.10

        :returns: a string with an informational message about this test

        .. note ::
           When overriding this method, you should pay extra attention on how
           you use the :class:`RegressionTest`'s attributes, because this
           method may be called at any point of the test's lifetime.
        '''

        ret = f'{self.display_name} /{self.hashcode}'
        if self.current_partition:
            ret += f' @{self.current_partition.fullname}'

        if self.current_environ:
            ret += f'+{self.current_environ.name}'

        return ret

    def is_local(self):
        '''Check if the test will execute locally.

        A test executes locally if the :attr:`local` attribute is set or if the
        current partition's scheduler does not support job submission.
        '''
        if self._current_partition is None:
            return self.local

        return self.local or self._current_partition.scheduler.is_local

    def is_fixture(self):
        '''Check if the test is a fixture.'''
        return getattr(self, '_rfm_is_fixture', False)

    def _resolve_fixtures(self):
        '''Resolve the fixture dependencies and inject the fixture handle.

        The fixture handle will point directly to the fixture object when the
        associated fixture uses a 'fork' action. However, when a fixture uses
        the 'join' action, the injected handle will point to a list with all
        the available fixture variants.
        '''

        current_part = self.current_partition.fullname
        current_env = self.current_environ.name

        # Get the declared and registered fixtures
        fixtures = type(self)._rfm_fixture_space
        registry = getattr(self, '_rfm_fixture_registry', None)

        # If this instance does not have a variant number, return
        if self.fixture_variant is None:
            return

        # Get the fixture variants required for this test variant.
        # This would retrieve the same information than when calling
        # ``type(self).get_variant_info(self.variant_num)['fixtures']``.
        target_fixt_variants = type(self).fixture_space[self.fixture_variant]

        for handle_name, f in fixtures.items():
            # Get the target variants for this fixture
            target_variants = target_fixt_variants[handle_name]

            # Prepare the getdep argumens based on the fixture's scope
            if f.scope == 'session':
                part = '*'
                environ = '*'
            elif f.scope == 'partition':
                part = None
                environ = '*'
            else:
                part = None
                environ = None

            # List to store all the targeted fixture variants
            deps = []

            # Scan the fixture registry and resolve the fixtures.
            # NOTE: The fixture registry can have multiple fixture instances
            # registered under the same fixture class. So the loop below must
            # also inspect the fixture data the instance was registered with.
            for fixt_name, fixt_data in registry[f.cls].items():
                if f.scope != fixt_data.scope:
                    continue
                elif fixt_data.variant_num not in target_variants:
                    continue
                elif f.scope == 'partition':
                    if fixt_data.partitions[0] != current_part:
                        continue
                elif f.scope == 'environment':
                    if (fixt_data.environments[0] != current_env or
                        fixt_data.partitions[0] != current_part):
                        continue

                # Resolve the fixture
                deps.append(self.getdep(fixt_name, environ, part))

            if f.action == 'fork':
                # When using the fork action, a fixture handle can only have
                # a single fixture instance attached. This could only happen
                # if either any of the above ifs is buggy or if a fixture with
                # a fork action ever has more than one index per fork variant.
                # None of this can happen from user input, but this must stay
                # here to ensure the unit tests do not fail silently.
                if len(deps) > 1:
                    raise PipelineError(
                        f'fixture {handle_name!r} has more than one instances'
                    )

                deps = deps[0]

            # Inject the fixtures
            setattr(self, handle_name, deps)

    def _setup_paths(self):
        '''Setup the check's dynamic paths.'''
        self.logger.debug('Setting up test paths')
        try:
            runtime = rt.runtime()
            self._stagedir = runtime.make_stagedir(
                self.current_system.name, self._current_partition.name,
                self._current_environ.name, self.short_name
            )
            self._outputdir = runtime.make_outputdir(
                self.current_system.name, self._current_partition.name,
                self._current_environ.name, self.short_name
            )
        except OSError as e:
            raise PipelineError('failed to set up paths') from e

    def _create_job(self, name, force_local=False, **job_opts):
        '''Setup the job related to this check.'''

        if force_local:
            scheduler = getscheduler('local')()
            launcher = getlauncher('local')()
        else:
            scheduler = self._current_partition.scheduler
            launcher = self._current_partition.launcher_type()

        self.logger.debug(
            f'Setting up job {name!r} '
            f'(scheduler: {scheduler.registered_name!r}, '
            f'launcher: {launcher.registered_name!r})'
        )
        return Job.create(scheduler,
                          launcher,
                          name=name,
                          workdir=self._stagedir,
                          sched_access=self._current_partition.access,
                          **job_opts)

    def _setup_build_job(self, **job_opts):
        self._build_job = self._create_job(f'rfm_build',
                                           self.local or self.build_locally,
                                           **job_opts)

    def _setup_run_job(self, **job_opts):
        self._job = self._create_job(f'rfm_job', self.local, **job_opts)

    def _setup_container_platform(self):
        try:
            self.container_platform.emit_prepare_commands(self.stagedir)
        except NotImplementedError:
            cplatf_name = self.current_partition.container_runtime
            if cplatf_name:
                try:
                    self.container_platform = ContainerPlatform.create_from(
                        cplatf_name, self.container_platform
                    )
                except ValueError:
                    pass

    @final
    def setup(self, partition, environ, **job_opts):
        '''The setup phase of the regression test pipeline.

        :arg partition: The system partition to set up this test for.
        :arg environ: The environment to set up this test for.
        :arg job_opts: Options to be passed through to the backend scheduler.
            When overriding this method users should always pass through
            ``job_opts`` to the base class method.
        :raises reframe.core.exceptions.ReframeError: In case of errors.

        .. warning::

           .. versionchanged:: 3.0
              You may not override this method directly unless you are in
              special test. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

           .. versionchanged:: 3.4
              Overriding this method directly in no longer allowed. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

        '''
        self._current_partition = partition
        self._current_environ = environ
        self._setup_paths()
        self._setup_build_job(**job_opts)
        self._setup_run_job(**job_opts)
        self._setup_container_platform()
        self._resolve_fixtures()

    def _copy_to_stagedir(self, path):
        self.logger.debug(f'Copying {path} to stage directory')
        self.logger.debug(f'Symlinking files: {self.readonly_files}')
        try:
            osext.copytree_virtual(
                path, self._stagedir, self.readonly_files, dirs_exist_ok=True
            )
        except (OSError, ValueError, TypeError) as e:
            raise PipelineError('copying of files failed') from e

    def _clone_to_stagedir(self, url):
        self.logger.debug(f'Cloning URL {url} into stage directory')
        osext.git_clone(
            self.sourcesdir, self._stagedir,
            timeout=rt.runtime().get_option('general/0/git_timeout')
        )

    @final
    def compile(self):
        '''The compilation phase of the regression test pipeline.

        :raises reframe.core.exceptions.ReframeError: In case of errors.

        .. warning::

           .. versionchanged:: 3.0
              You may not override this method directly unless you are in
              special test. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

           .. versionchanged:: 3.4
              Overriding this method directly in no longer allowed. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

        '''

        # Copy the check's resources to the stage directory
        if self.sourcesdir:
            try:
                commonpath = os.path.commonpath([self.sourcesdir,
                                                 self.sourcepath])
            except ValueError:
                commonpath = None

            if commonpath:
                self.logger.warn(
                    f'sourcepath {self.sourcepath!r} is a subdirectory of '
                    f'sourcesdir {self.sourcesdir!r}, but it will be '
                    f'interpreted as relative to it'
                )

            if osext.is_url(self.sourcesdir):
                self._clone_to_stagedir(self.sourcesdir)
            else:
                self._copy_to_stagedir(os.path.join(self._prefix,
                                                    self.sourcesdir))

        # Set executable (only if hasn't been provided)
        if not hasattr(self, 'executable'):
            self.executable = os.path.join('.', self.unique_name)

        # Verify the sourcepath and determine the sourcepath in the stagedir
        if (os.path.isabs(self.sourcepath) or
                os.path.normpath(self.sourcepath).startswith('..')):
            raise PipelineError(
                'self.sourcepath is an absolute path or does not point to a '
                'subfolder or a file contained in self.sourcesdir: ' +
                self.sourcepath
            )

        staged_sourcepath = os.path.join(self._stagedir, self.sourcepath)
        if os.path.isdir(staged_sourcepath):
            if not self.build_system:
                # Try to guess the build system
                cmakelists = os.path.join(staged_sourcepath, 'CMakeLists.txt')
                configure_ac = os.path.join(staged_sourcepath, 'configure.ac')
                configure_in = os.path.join(staged_sourcepath, 'configure.in')
                if os.path.exists(cmakelists):
                    self.build_system = 'CMake'
                    self.build_system.builddir = 'rfm_build'
                elif (os.path.exists(configure_ac) or
                      os.path.exists(configure_in)):
                    self.build_system = 'Autotools'
                    self.build_system.builddir = 'rfm_build'
                else:
                    self.build_system = 'Make'

            self.build_system.srcdir = self.sourcepath
        else:
            if not self.build_system:
                self.build_system = 'SingleSource'

            self.build_system.srcfile = self.sourcepath
            self.build_system.executable = self.executable

        user_environ = Environment(self.unique_name,
                                   self.modules, self.env_vars.items())
        environs = [self._current_partition.local_env, self._current_environ,
                    user_environ, self._cdt_environ]
        self._build_job.time_limit = (
            self.build_time_limit or rt.runtime().get_option(
                f'systems/0/partitions/@{self.current_partition.name}'
                f'/time_limit')
        )
        # Get job options from managed resources and prepend them to
        # build_job_opts. We want any user supplied options to be able to
        # override those set by the framework.
        resources_opts = self._map_resources_to_jobopts()
        self._build_job.options = resources_opts + self._build_job.options
        with osext.change_dir(self._stagedir):
            # Prepare build job
            build_commands = [
                *self.prebuild_cmds,
                *self.build_system.emit_build_commands(self._current_environ),
                *self.postbuild_cmds
            ]
            try:
                self._build_job.prepare(
                    build_commands, environs,
                    self._current_partition.prepare_cmds,
                    login=rt.runtime().get_option('general/0/use_login_shell'),
                    trap_errors=True
                )
            except OSError as e:
                raise PipelineError('failed to prepare build job') from e

            self._build_job.submit()

    @final
    def compile_wait(self):
        '''Wait for compilation phase to finish.

        .. versionadded:: 2.13

        .. warning::

           .. versionchanged:: 3.0
              You may not override this method directly unless you are in
              special test. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

           .. versionchanged:: 3.4
              Overriding this method directly in no longer allowed. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

        '''
        self._build_job.wait()

        # We raise a BuildError when we an exit code and it is non zero
        if self._build_job.exitcode:
            raise BuildError(self._build_job.stdout,
                             self._build_job.stderr, self._stagedir)

        with osext.change_dir(self._stagedir):
            self.build_system.post_build(self._build_job)

    @final
    def run(self):
        '''The run phase of the regression test pipeline.

        This call is non-blocking.
        It simply submits the job associated with this test and returns.

        .. warning::

           .. versionchanged:: 3.0
              You may not override this method directly unless you are in
              special test. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

           .. versionchanged:: 3.4
              Overriding this method directly in no longer allowed. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

        '''

        def _get_cp_env():
            '''Retrieve the container platform environment.'''
            try:
                cp_name = self.container_platform.name
                return self.current_partition.container_environs[cp_name]
            except KeyError:
                return None

        cp = self.container_platform
        if cp.image:
            # We replace executable and executable_opts in case of containers
            try:
                self.executable = cp.launch_command(self.stagedir)
                self.executable_opts = []
                prepare_container = cp.emit_prepare_commands(self.stagedir)
                if prepare_container:
                    self.prerun_cmds += prepare_container
            except NotImplementedError:
                raise PipelineError(
                    "no container runtime was configured; "
                    "consider setting the 'container_platform' test attribute "
                    "or the corresponding partition configuration setting"
                )

        self.job.num_tasks = self.num_tasks
        self.job.num_tasks_per_node = self.num_tasks_per_node
        self.job.num_tasks_per_core = self.num_tasks_per_core
        self.job.num_tasks_per_socket = self.num_tasks_per_socket
        self.job.num_cpus_per_task = self.num_cpus_per_task
        self.job.use_smt = self.use_multithreading
        self.job.time_limit = (self.time_limit or rt.runtime().get_option(
            f'systems/0/partitions/@{self.current_partition.name}/time_limit')
        )
        self.job.max_pending_time = self.max_pending_time
        self.job.exclusive_access = self.exclusive_access
        exec_cmd = [self.job.launcher.run_command(self.job),
                    self.executable, *self.executable_opts]

        if self.build_system:
            prepare_cmds = self.build_system.prepare_cmds()
        else:
            prepare_cmds = []

        commands = [
            *prepare_cmds,
            *self.prerun_cmds,
            ' '.join(exec_cmd).strip(),
            *self.postrun_cmds
        ]
        user_environ = Environment(self.unique_name,
                                   self.modules, self.env_vars.items())
        environs = [
            self._current_partition.local_env,
            self._current_environ,
            user_environ,
            self._cdt_environ
        ]
        if self.container_platform.image:
            cp_env = _get_cp_env()
            if cp_env:
                environs.insert(2, cp_env)

        # num_gpus_per_node is a managed resource
        if self.num_gpus_per_node:
            self.extra_resources.setdefault(
                '_rfm_gpu', {'num_gpus_per_node': self.num_gpus_per_node}
            )

        # Get job options from managed resources and prepend them to
        # job_opts. We want any user supplied options to be able to
        # override those set by the framework.
        resources_opts = self._map_resources_to_jobopts()
        self._job.options = resources_opts + self._job.options
        with osext.change_dir(self._stagedir):
            try:
                self.logger.debug('Generating the run script')
                self._job.prepare(
                    commands, environs,
                    self._current_partition.prepare_cmds,
                    login=rt.runtime().get_option('general/0/use_login_shell'),
                    trap_errors=rt.runtime().get_option(
                        'general/0/trap_job_errors'
                    )
                )
            except OSError as e:
                raise PipelineError('failed to prepare run job') from e

            self._job.submit()

        self.logger.debug(f'Spawned run job (id={self.job.jobid})')

        # Update num_tasks if test is flexible
        if self.job.sched_flex_alloc_nodes:
            self.num_tasks = self.job.num_tasks

    def _map_resources_to_jobopts(self):
        resources_opts = []
        for r, v in self.extra_resources.items():
            resources_opts += self._current_partition.get_resource(r, **v)

        return resources_opts

    @final
    def compile_complete(self):
        '''Check if the build phase has completed.

        :returns: :class:`True` if the associated build job has finished,
            :class:`False` otherwise.

            If no job descriptor is yet associated with this test,
            :class:`True` is returned.
        :raises reframe.core.exceptions.ReframeError: In case of errors.

        '''
        if not self._build_job:
            return True

        return self._build_job.finished()

    @final
    def run_complete(self):
        '''Check if the run phase has completed.

        :returns: :class:`True` if the associated job has finished,
            :class:`False` otherwise.

            If no job descriptor is yet associated with this test,
            :class:`True` is returned.
        :raises reframe.core.exceptions.ReframeError: In case of errors.

        .. warning::
           You may not override this method directly unless you are in
           special test. See `here
           <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
           more details.


           .. versionchanged:: 3.4
              Overriding this method directly in no longer allowed. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

        '''
        if not self._job:
            return True

        return self._job.finished()

    @final
    def run_wait(self):
        '''Wait for the run phase of this test to finish.

        :raises reframe.core.exceptions.ReframeError: In case of errors.

        .. warning::
           You may not override this method directly unless you are in
           special test. See `here
           <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
           more details.

           .. versionchanged:: 3.4
              Overriding this method directly in no longer allowed. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

        '''
        self._job.wait()

    @final
    def sanity(self):
        self.check_sanity()

    @final
    def performance(self):
        try:
            self.check_performance()
        except PerformanceError:
            if self.strict_check:
                raise

    @final
    def check_sanity(self):
        '''The sanity checking phase of the regression test pipeline.

        :raises reframe.core.exceptions.SanityError: If the sanity check fails.
        :raises reframe.core.exceptions.ReframeSyntaxError: If the sanity
            function cannot be resolved due to ambiguous syntax.

        .. warning::

           .. versionchanged:: 3.0
              You may not override this method directly unless you are in
              special test. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

           .. versionchanged:: 3.4
              Overriding this method directly in no longer allowed. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

        '''

        if hasattr(self, '_rfm_sanity'):
            # Using more than one type of syntax to set the sanity patterns is
            # not allowed.
            if hasattr(self, 'sanity_patterns'):
                raise ReframeSyntaxError(
                    f"assigning a sanity function to the 'sanity_patterns' "
                    f"variable conflicts with using the 'sanity_function' "
                    f"decorator (class {self.__class__.__qualname__})"
                )

            self.sanity_patterns = self._rfm_sanity()

        if rt.runtime().get_option('general/0/trap_job_errors'):
            sanity_patterns = [
                sn.assert_eq(self.job.exitcode, 0,
                             msg='job exited with exit code {0}')
            ]
            if hasattr(self, 'sanity_patterns'):
                sanity_patterns.append(self.sanity_patterns)

            self.sanity_patterns = sn.all(sanity_patterns)
        elif not hasattr(self, 'sanity_patterns'):
            raise SanityError('sanity_patterns not set')

        with osext.change_dir(self._stagedir):
            success = sn.evaluate(self.sanity_patterns)
            if not success:
                raise SanityError()

    def is_performance_check(self):
        '''Return :obj:`True` if the test is a performance test.'''
        return self.perf_variables or hasattr(self, 'perf_patterns')

    @final
    def check_performance(self):
        '''The performance checking phase of the regression test pipeline.

        :raises reframe.core.exceptions.SanityError: If the performance check
            fails.

        .. warning::

           .. versionchanged:: 3.0
              You may not override this method directly unless you are in
              special test. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

           .. versionchanged:: 3.4
              Overriding this method directly in no longer allowed. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

        '''

        if not self.is_performance_check():
            return

        perf_patterns = getattr(self, 'perf_patterns', None)

        if perf_patterns is not None and self.perf_variables:
            raise ReframeSyntaxError(
                f"you cannot mix 'perf_patterns' and 'perf_variables' syntax"
            )

        # Convert `perf_patterns` to `perf_variables`
        if perf_patterns:
            for var, expr in self.perf_patterns.items():
                # Retrieve the unit from the reference tuple
                key = f'{self._current_partition.fullname}:{var}'
                try:
                    unit = self.reference[key][3]
                    if unit is None:
                        unit = ''
                except KeyError:
                    unit = ''

                self.perf_variables[var] = sn.make_performance_function(expr,
                                                                        unit)

        # Evaluate the performance function and retrieve the metrics
        with osext.change_dir(self._stagedir):
            for tag, expr in self.perf_variables.items():
                try:
                    value = expr.evaluate()
                    unit = expr.unit
                except Exception as e:
                    logging.getlogger().warning(
                        f'skipping evaluation of performance variable '
                        f'{tag!r}: {e}'
                    )
                    continue

                key = f'{self._current_partition.fullname}:{tag}'
                try:
                    ref = self.reference[key]

                    # If units are also provided in the reference, raise
                    # a warning if they match with the units provided by
                    # the performance function.
                    if len(ref) == 4:
                        if ref[3] != unit:
                            logging.getlogger().warning(
                                f'reference unit ({key!r}) for the '
                                f'performance variable {tag!r} '
                                f'does not match the unit specified '
                                f'in the performance function ({unit!r}): '
                                f'{unit!r} will be used'
                            )

                        # Pop the unit from the ref tuple (redundant)
                        ref = ref[:3]
                except KeyError:
                    if self.require_reference:
                        raise PerformanceError(
                            f'no reference value found for '
                            f'performance variable {tag!r} on '
                            f'system {self._current_partition.fullname!r}'
                        ) from None

                    ref = (0, None, None)

                self._perfvalues[key] = (value, *ref, unit)

            # Check the performance variables against their references.
            for key, values in self._perfvalues.items():
                val, ref, low_thres, high_thres, *_ = values

                # Verify that val is a number
                if not isinstance(val, numbers.Number):
                    raise SanityError(
                        f'the value extracted for performance variable '
                        f'{key!r} is not a number: {val}'
                    )

                tag = key.split(':')[-1]
                try:
                    sn.evaluate(
                        sn.assert_reference(
                            val, ref, low_thres, high_thres,
                            msg=('failed to meet reference: %s={0}, '
                                 'expected {1} (l={2}, u={3})' % tag))
                    )
                except SanityError as e:
                    raise PerformanceError(e) from None

    def _copy_job_files(self, job, dst):
        if job is None:
            return

        stdout = os.path.join(self._stagedir, job.stdout)
        stderr = os.path.join(self._stagedir, job.stderr)
        script = os.path.join(self._stagedir, job.script_filename)
        shutil.copy(stdout, dst)
        shutil.copy(stderr, dst)
        shutil.copy(script, dst)

    def _copy_to_outputdir(self):
        '''Copy check's interesting files to the output directory.'''
        self.logger.debug('Copying test files to output directory')
        self._copy_job_files(self._job, self.outputdir)
        self._copy_job_files(self._build_job, self.outputdir)

        with osext.change_dir(self.stagedir):
            # Copy files specified by the user, but expand any glob patterns
            keep_files = itertools.chain(
                *(glob.iglob(f) for f in self.keep_files)
            )
            for f in keep_files:
                f = os.path.abspath(f)
                if os.path.isdir(f):
                    # We need to keep the directory structure when copying
                    # over to outputdir
                    dst = os.path.join(
                        self.outputdir, os.path.relpath(f, self.stagedir)
                    )
                    osext.copytree(f, dst, dirs_exist_ok=True)
                else:
                    shutil.copy2(f, self.outputdir)

    @final
    def cleanup(self, remove_files=False):
        '''The cleanup phase of the regression test pipeline.

        :arg remove_files: If :class:`True`, the stage directory associated
            with this test will be removed.

        .. warning::

           .. versionchanged:: 3.0
              You may not override this method directly unless you are in
              special test. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

           .. versionchanged:: 3.4
              Overriding this method directly in no longer allowed. See `here
              <migration_2_to_3.html#force-override-a-pipeline-method>`__ for
              more details.

        '''
        aliased = os.path.samefile(self._stagedir, self._outputdir)
        if aliased:
            self.logger.debug(
                f'outputdir and stagedir are the same; copying skipped'
            )
        else:
            self._copy_to_outputdir()

        if remove_files:
            self.logger.debug('Removing stage directory')
            osext.rmtree(self._stagedir)

    # Dependency API

    def user_deps(self):
        return util.SequenceView(self._userdeps)

    def depends_on(self, target, how=None, *args, **kwargs):
        '''Add a dependency to another test.

        :arg target: The name of the test that this one will depend on.
        :arg how: A callable that defines how the test cases of this test
            depend on the the test cases of the target test.
            This callable should accept two arguments:

            - The source test case (i.e., a test case of this test)
              represented as a two-element tuple containing the names of the
              partition and the environment of the current test case.
            - Test destination test case (i.e., a test case of the target
              test) represented as a two-element tuple containing the names of
              the partition and the environment of the current target test
              case.

            It should return :class:`True` if a dependency between the source
            and destination test cases exists, :class:`False` otherwise.

            This function will be called multiple times by the framework when
            the test DAG is constructed, in order to determine the
            connectivity of the two tests.

            In the following example, this test depends on ``T1`` when their
            partitions match, otherwise their test cases are independent.

            .. code-block:: python

                def by_part(src, dst):
                    p0, _ = src
                    p1, _  = dst
                    return p0 == p1

                self.depends_on('T0', how=by_part)

            The framework offers already a set of predefined relations between
            the test cases of inter-dependent tests. See the
            :mod:`reframe.utility.udeps` for more details.

            The default ``how`` function is
            :func:`reframe.utility.udeps.by_case`, where test cases on
            different partitions and environments are independent.

        .. seealso::
           - :doc:`dependencies`
           - :ref:`test-case-deps-management`



        .. versionadded:: 2.21

        .. versionchanged:: 3.3
           Dependencies between test cases from different partitions are now
           allowed. The ``how`` argument now accepts a callable.

         .. deprecated:: 3.3
            Passing an integer to the ``how`` argument as well as using the
            ``subdeps`` argument is deprecated.

        .. versionchanged:: 4.0.0
           Passing an integer to the ``how`` argument is no longer supported.

        '''
        if not isinstance(target, str):
            raise TypeError("target argument must be of type: `str'")

        if how is None:
            how = udeps.by_case

        if not callable(how):
            raise TypeError("'how' argument must be callable")

        self._userdeps.append((target, how))

    def getdep(self, target, environ=None, part=None):
        '''Retrieve the test case of a target dependency.

        This is a low-level method. The :func:`@require_deps
        <reframe.core.decorators.require_deps>` decorators should be
        preferred.

        :arg target: The name of the target dependency to be retrieved.
        :arg environ: The name of the programming environment that will be
            used to retrieve the test case of the target test. If ``None``,
            :attr:`RegressionTest.current_environ` will be used.

        .. versionadded:: 2.21

        .. versionchanged:: 3.8.0
           Setting ``environ`` or ``part`` to ``'*'`` will skip the match
           check on the environment and partition, respectively.

        '''
        if self.current_environ is None:
            raise DependencyError(
                'cannot resolve dependencies before the setup phase'
            )

        if environ is None:
            environ = self.current_environ.name

        if part is None:
            part = self.current_partition.name

        if self._case is None or self._case() is None:
            raise DependencyError('no test case is associated with this test')

        for d in self._case().deps:
            mask = int(d.check.unique_name == target)
            mask |= (int(d.partition.name == part) | int(part == '*')) << 1
            mask |= (int(d.environ.name == environ) | int(environ == '*')) << 2
            if mask == 7:
                return d.check

        raise DependencyError(f'could not resolve dependency to ({target!r}, '
                              f'{part!r}, {environ!r})')

    def skip(self, msg=None):
        '''Skip test.

        :arg msg: A message explaining why the test was skipped.

        .. versionadded:: 3.5.1
        '''
        raise SkipTestError(msg)

    def skip_if(self, cond, msg=None):
        '''Skip test if condition is true.

        :arg cond: The condition to check for skipping the test.
        :arg msg: A message explaining why the test was skipped.

        .. versionadded:: 3.5.1
        '''
        if cond:
            self.skip(msg)

    def skip_if_no_procinfo(self, msg=None):
        '''Skip test if no processor topology information is available.

        This method has effect only if called after the ``setup`` stage.

        :arg msg: A message explaining why the test was skipped.
            If not specified, a default message will be used.

        .. versionadded:: 3.9.1
        '''
        if not self.current_partition:
            return

        proc = self.current_partition.processor
        pname = self.current_partition.fullname
        if msg is None:
            msg = f'no topology information found for partition {pname!r}'

        self.skip_if(not proc.info, msg)

    def __str__(self):
        return f'{self.unique_name} [{self.display_name}]'

    def __eq__(self, other):
        if not isinstance(other, RegressionTest):
            return NotImplemented

        return self.unique_name == other.unique_name

    def __hash__(self):
        return hash(self.unique_name)

    def __rfm_json_decode__(self, json):
        # 'tags' are decoded as list, so we convert them to a set
        self.tags = set(json['tags'])


class RunOnlyRegressionTest(RegressionTest, special=True):
    '''Base class for run-only regression tests.

    This class is also directly available under the top-level :mod:`reframe`
    module.
    '''

    _rfm_regression_class_kind = _RFM_TEST_KIND_RUN

    def setup(self, partition, environ, **job_opts):
        '''The setup stage of the regression test pipeline.

        Similar to the :func:`RegressionTest.setup`, except that no build job
        is created for this test.
        '''
        self._current_partition = partition
        self._current_environ = environ
        self._setup_paths()
        self._setup_run_job(**job_opts)
        self._setup_container_platform()
        self._resolve_fixtures()

    def compile(self):
        '''The compilation phase of the regression test pipeline.

        This is a no-op for this type of test.
        '''

    def compile_wait(self):
        '''Wait for compilation phase to finish.

        This is a no-op for this type of test.
        '''

    def run(self):
        '''The run phase of the regression test pipeline.

        The resources of the test are copied to the stage directory and the
        rest of execution is delegated to the :func:`RegressionTest.run()`.
        '''
        if self.sourcesdir:
            if osext.is_url(self.sourcesdir):
                self._clone_to_stagedir(self.sourcesdir)
            else:
                self._copy_to_stagedir(os.path.join(self._prefix,
                                                    self.sourcesdir))

        super().run()


class CompileOnlyRegressionTest(RegressionTest, special=True):
    '''Base class for compile-only regression tests.

    These tests are by default local and will skip the run phase of the
    regression test pipeline.

    The standard output and standard error of the test will be set to those of
    the compilation stage.

    This class is also directly available under the top-level :mod:`reframe`
    module.
    '''

    _rfm_regression_class_kind = _RFM_TEST_KIND_COMPILE

    def setup(self, partition, environ, **job_opts):
        '''The setup stage of the regression test pipeline.

        Similar to the :func:`RegressionTest.setup`, except that no run job
        is created for this test.
        '''
        # No need to setup the job for compile-only checks
        self._current_partition = partition
        self._current_environ = environ
        self._setup_paths()
        self._setup_build_job(**job_opts)
        self._setup_container_platform()
        self._resolve_fixtures()

    @property
    @deferrable
    def stdout(self):
        return self.build_job.stdout if self.build_job else None

    @property
    @deferrable
    def stderr(self):
        return self.build_job.stderr if self.build_job else None

    def run(self):
        '''The run stage of the regression test pipeline.

        Implemented as no-op.
        '''

    def run_wait(self):
        '''Wait for this test to finish.

        Implemented as no-op
        '''



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

#
# LSF backend
#
# - Initial version submitted by Ryan Goodner, UNM (based on PBS backend)
# - Extended and fixed by Jonathan Frawley and Mark Turner
#

import functools
import re
import time

import reframe.utility.osext as osext
from reframe.core.backends import register_scheduler
from reframe.core.exceptions import JobSchedulerError
from reframe.core.schedulers.pbs import PbsJobScheduler

_run_strict = functools.partial(osext.run_command, check=True)


@register_scheduler('lsf')
class LsfJobScheduler(PbsJobScheduler):
    def __init__(self):
        self._prefix = '#BSUB'
        self._submit_timeout = self.get_option('job_submit_timeout')

    def _format_option(self, var, option):
        if var is not None:
            return self._prefix + ' ' + option.format(var)
        else:
            return ''

    def emit_preamble(self, job):
        preamble = [
            self._format_option(job.name, '-J {0}'),
            self._format_option(job.stdout, '-o {0}'),
            self._format_option(job.stderr, '-e {0}'),
        ]

        if job.num_tasks_per_node is not None:
            num_nodes = job.num_tasks // job.num_tasks_per_node
            preamble.append(self._format_option(num_nodes, '-nnodes {0}'))
        else:
            preamble.append(self._format_option(job.num_tasks, '-n {0}'))

        if job.num_cpus_per_task is not None:
            preamble.append(self._format_option(job.num_cpus_per_task,
                                                '-R "affinity[core({0})]"'))

        # add job time limit in minutes
        if job.time_limit is not None:
            preamble.append(
                f'{self._prefix} -W {int(job.time_limit // 60)}'
            )

        for opt in job.sched_access:
            preamble.append(f'{self._prefix} {opt}')

        # emit the rest of the options
        options = job.options + job.cli_options
        for opt in options:
            if opt.startswith('#'):
                preamble.append(opt)
            else:
                preamble.append(self._prefix + ' ' + opt)

        if job.exclusive_access:
            preamble.append(f'{self._prefix} -x')

        # Filter out empty statements before returning
        return list(filter(None, preamble))

    def submit(self, job):
        with open(job.script_filename, 'r') as fp:
            completed = _run_strict('bsub', stdin=fp)
        jobid_match = re.search(r'^Job <(?P<jobid>\S+)> is submitted',
                                completed.stdout)
        if not jobid_match:
            raise JobSchedulerError('could not retrieve the job id '
                                    'of the submitted job')
        job._jobid = jobid_match.group('jobid')
        job._submit_time = time.time()

    def poll(self, *jobs):
        if jobs:
            # filter out non-jobs
            jobs = [job for job in jobs if job is not None]

        if not jobs:
            return

        completed = _run_strict(
            f'bjobs -noheader {" ".join(job.jobid for job in jobs)}'
        )
        job_status = {}
        job_status_lines = completed.stdout.split('\n')

        for line in job_status_lines:
            job_regex = (r'(?P<jobid>\d+)\s+'
                         r'(?P<user>\S+)\s+'
                         r'(?P<status>\S+)\s+'
                         r'(?P<queue>\S+)')
            job_match = re.search(job_regex, line)
            if job_match:
                job_status[job_match['jobid']] = job_match['status']

        for job in jobs:
            if job.jobid not in job_status:
                # job id not found
                self.log(f'Job {job.jobid} not known to scheduler, '
                         f'assuming job completed')
                job._state = 'COMPLETED'
                job._completed = True
            elif job_status[job.jobid] in ('DONE', 'EXIT'):
                # job done
                job._state = 'COMPLETED'
                job._completed = True
            elif job_status[job.jobid] == 'RUN':
                # job running
                job._state = 'RUNNING'
            elif job_status[job.jobid] == 'PEND':
                # job pending
                job._state = 'PENDING'
            elif job_status[job.jobid] in ['PSUSP', 'SSUSP', 'USUSP']:
                # job suspended
                job._state = 'SUSPENDED'
            else:
                # job status unknown
                self.log(f'Job {job_status[job.jobid]} not known, '
                         f'assuming job completed')
                job._state = 'COMPLETED'
                job._completed = True

    def finished(self, job):
        if job.exception:
            raise job.exception

        return job.state == 'COMPLETED'



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

#
# Base classes to manage the namespace of a regression test.
#


import abc

from reframe.core.exceptions import ReframeSyntaxError


class LocalNamespace:
    '''Local namespace of a regression test.

    Temporary storage for test attributes defined in the test class body.
    This local namespace is populated during the test class body execution.

    Example: In the pseudo-code below, the local namespace of A is {P0},
    and the local namespace of B is {P1}. However, the final namespace
    of A is still {P0}, and the final namespace of B is {P0, P1}.

    .. code:: python

        class A(RegressionTest):
            var('P0')

        class B(A):
            var('P1')
    '''

    def __init__(self, namespace=None):
        self._namespace = namespace or {}

    def __getattr__(self, name):
        return getattr(self._namespace, name)

    def __getitem__(self, key):
        return self._namespace[key]

    def __setitem__(self, key, value):
        if key not in self._namespace:
            self._namespace[key] = value
        else:
            self._raise_namespace_clash(key)

    def __contains__(self, key):
        return key in self._namespace

    def __iter__(self):
        return iter(self._namespace)

    def __len__(self):
        return len(self._namespace)

    def __repr__(self):
        return f'{type(self).__name__}({self._namespace!r})'

    def _raise_namespace_clash(self, name):
        '''Raise an error if there is a namespace clash.'''
        raise KeyError(
            f'{name!r} is already present in the current namespace'
        )

    def clear(self):
        self._namespace = {}

    def data(self):
        '''Give access to the underlying namespace'''
        return self._namespace


class Namespace(LocalNamespace, metaclass=abc.ABCMeta):
    '''Namespace of a regression test.

    The final namespace may be built by inheriting namespaces from the base
    classes, and extending this one with the information stored in the local
    namespace of the target class. In this context, the target class is simply
    the regression test class where the namespace is to be built.

    If a target class is provided, the constructor will build a Namespace
    instance by inheriting the namespaces found in the base classes, and
    extending this with the information from the local namespace of the
    target class.

    Eventually, the items from a Namespace are injected as attributes of
    the target class instance by the :func:`inject` method, which must be
    called by the target class during its instantiation process. Also, a target
    class may use more that one Namespace, which raises the need for name
    checking across namespaces. Thus, the :func:`__init__` method accepts the
    additional argument ``illegal_names``, which is a set of class attribute
    names already in use by the target class or other namespaces from this
    target class. Then, after the Namespace is built, if ``illegal_names`` is
    provided, a sanity check is performed, ensuring that no name clashing
    will occur during the target class instantiation process.

    '''

    def __init__(self, target_cls=None, illegal_names=None,
                 *, ns_name, ns_local_name):
        super().__init__()
        self._ns_name = ns_name
        self._ns_local_name = ns_local_name
        if target_cls:
            # Inherit Namespaces from the base clases
            self.inherit(target_cls)

            # Extend the Namespace with the LocalNamespace
            self.extend(target_cls)

            # Sanity checkings on the resulting Namespace
            self.sanity(target_cls, illegal_names)

    @property
    def namespace_name(self):
        return self._ns_name

    @property
    def local_namespace_name(self):
        return self._ns_local_name

    def inherit(self, cls):
        '''Inherit the Namespaces from the bases.'''

        for base in cls.__bases__:
            other = getattr(base, self.namespace_name, None)
            if isinstance(other, type(self)):
                self.join(other, cls)

    @abc.abstractmethod
    def join(self, other, cls):
        '''Join other Namespace with the current one.'''

    @abc.abstractmethod
    def extend(self, cls):
        '''Extend the namespace with the local namespace.'''

    def sanity(self, cls, illegal_names):
        '''Sanity checks post-creation of the namespace.

        By default, we make illegal to have any item in the namespace
        that clashes with a member of the target class.
        '''
        if illegal_names is None:
            illegal_names = set(dir(cls))

        for key in self._namespace:
            if key in illegal_names:
                raise ReframeSyntaxError(
                    f'{key!r} already defined in class '
                    f'{cls.__qualname__!r}'
                )

    @abc.abstractmethod
    def inject(self, obj, objtype=None):
        '''Insert the items from the namespace as attributes of the object
           ``obj``.
        '''

    def __setitem__(self, key, value):
        raise ReframeSyntaxError(
            f'cannot set item {key!r} into a {type(self).__qualname__} object'
        )



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import contextlib
import math
import sys
import time

import reframe.core.runtime as rt
import reframe.utility as util
from reframe.core.exceptions import (FailureLimitError,
                                     SkipTestError,
                                     TaskDependencyError,
                                     TaskExit)
from reframe.core.logging import getlogger, level_from_str
from reframe.core.pipeline import (CompileOnlyRegressionTest,
                                   RunOnlyRegressionTest)
from reframe.frontend.executors import (ExecutionPolicy, RegressionTask,
                                        TaskEventListener, ABORT_REASONS)


def _get_partition_name(task, phase='run'):
    if (task.check.local or
        (phase == 'build' and task.check.build_locally)):
        return '_rfm_local'
    else:
        return task.check.current_partition.fullname


def _cleanup_all(tasks, *args, **kwargs):
    for task in tasks:
        if task.ref_count == 0:
            with contextlib.suppress(TaskExit):
                task.cleanup(*args, **kwargs)

    # Remove cleaned up tests
    tasks[:] = [t for t in tasks if t.ref_count]


def _print_perf(task):
    '''Get performance info of the current task.'''

    perfvars = task.testcase.check.perfvalues
    level = level_from_str(
        rt.runtime().get_option('general/0/perf_info_level')
    )
    for key, info in perfvars.items():
        name = key.split(':')[-1]
        getlogger().log(level,
                        f'P: {name}: {info[0]} {info[4]} '
                        f'(r:{info[1]}, l:{info[2]}, u:{info[3]})')


class _PollController:
    SLEEP_MIN = 0.1
    SLEEP_MAX = 10
    SLEEP_INC_RATE = 1.1

    def __init__(self):
        self._num_polls = 0
        self._sleep_duration = None
        self._t_init = None

    def reset_snooze_time(self):
        self._sleep_duration = self.SLEEP_MIN

    def snooze(self):
        if self._num_polls == 0:
            self._t_init = time.time()

        t_elapsed = time.time() - self._t_init
        self._num_polls += 1
        poll_rate = self._num_polls / t_elapsed if t_elapsed else math.inf
        getlogger().debug2(
            f'Poll rate control: sleeping for {self._sleep_duration}s '
            f'(current poll rate: {poll_rate} polls/s)'
        )
        time.sleep(self._sleep_duration)
        self._sleep_duration = min(
            self._sleep_duration*self.SLEEP_INC_RATE, self.SLEEP_MAX
        )


class SerialExecutionPolicy(ExecutionPolicy, TaskEventListener):
    def __init__(self):
        super().__init__()

        self._pollctl = _PollController()

        # Index tasks by test cases
        self._task_index = {}

        # Tasks that have finished, but have not performed their cleanup phase
        self._retired_tasks = []
        self.task_listeners.append(self)

    def runcase(self, case):
        super().runcase(case)
        check, partition, environ = case
        task = RegressionTask(case, self.task_listeners)
        self.printer.status('RUN', task.info())
        self._task_index[case] = task
        self.stats.add_task(task)
        try:
            # Do not run test if any of its dependencies has failed
            # NOTE: Restored dependencies are not in the task_index
            if any(self._task_index[c].failed
                   for c in case.deps if c in self._task_index):
                raise TaskDependencyError('dependencies failed')

            if any(self._task_index[c].skipped
                   for c in case.deps if c in self._task_index):

                # We raise the SkipTestError here and catch it immediately in
                # order for `skip()` to get the correct exception context.
                try:
                    raise SkipTestError('skipped due to skipped dependencies')
                except SkipTestError as e:
                    task.skip()
                    raise TaskExit from e

            task.setup(task.testcase.partition,
                       task.testcase.environ,
                       sched_flex_alloc_nodes=self.sched_flex_alloc_nodes,
                       sched_options=self.sched_options)

            task.compile()
            task.compile_wait()
            task.run()

            # Pick the right scheduler
            if task.check.local:
                sched = self.local_scheduler
            else:
                sched = partition.scheduler

            self._pollctl.reset_snooze_time()
            while True:
                sched.poll(task.check.job)
                if task.run_complete():
                    break

                self._pollctl.snooze()

            task.run_wait()
            if not self.skip_sanity_check:
                task.sanity()

            if not self.skip_performance_check:
                task.performance()

            self._retired_tasks.append(task)
            task.finalize()
        except TaskExit:
            return
        except ABORT_REASONS as e:
            task.abort(e)
            raise
        except BaseException:
            task.fail(sys.exc_info())

    def on_task_setup(self, task):
        pass

    def on_task_run(self, task):
        pass

    def on_task_compile(self, task):
        pass

    def on_task_exit(self, task):
        pass

    def on_task_compile_exit(self, task):
        pass

    def on_task_skip(self, task):
        msg = str(task.exc_info[1])
        self.printer.status('SKIP', msg, just='right')

    def on_task_failure(self, task):
        self._num_failed_tasks += 1
        msg = f'{task.info()}'
        if task.failed_stage == 'cleanup':
            self.printer.status('ERROR', msg, just='right')
        else:
            self.printer.status('FAIL', msg, just='right')

        _print_perf(task)
        timings = task.pipeline_timings(['setup',
                                         'compile_complete',
                                         'run_complete',
                                         'sanity',
                                         'performance',
                                         'total'])
        getlogger().info(f'==> test failed during {task.failed_stage!r}: '
                         f'test staged in {task.check.stagedir!r}')
        getlogger().verbose(f'==> {timings}')
        if self._num_failed_tasks >= self.max_failures:
            raise FailureLimitError(
                f'maximum number of failures ({self.max_failures}) reached'
            )

    def on_task_success(self, task):
        msg = f'{task.info()}'
        self.printer.status('OK', msg, just='right')
        _print_perf(task)
        timings = task.pipeline_timings(['setup',
                                         'compile_complete',
                                         'run_complete',
                                         'sanity',
                                         'performance',
                                         'total'])
        getlogger().verbose(f'==> {timings}')

        # Update reference count of dependencies
        for c in task.testcase.deps:
            # NOTE: Restored dependencies are not in the task_index
            if c in self._task_index:
                self._task_index[c].ref_count -= 1

        _cleanup_all(self._retired_tasks, not self.keep_stage_files)

    def exit(self):
        # Clean up all remaining tasks
        _cleanup_all(self._retired_tasks, not self.keep_stage_files)


class AsynchronousExecutionPolicy(ExecutionPolicy, TaskEventListener):
    '''The asynchronous execution policy.'''

    def __init__(self):
        super().__init__()

        self._pollctl = _PollController()

        # Index tasks by test cases
        self._task_index = {}

        # A set of all the current tasks. We use an ordered set here, because
        # we want to preserve the order of the tasks.
        self._current_tasks = util.OrderedSet()

        # Quick look up for the partition schedulers including the
        # `_rfm_local` pseudo-partition
        self._schedulers = {
            '_rfm_local': self.local_scheduler
        }

        # Tasks per partition
        self._partition_tasks = {
            '_rfm_local': util.OrderedSet()
        }

        # Retired tasks that need to be cleaned up
        self._retired_tasks = []

        # Job limit per partition
        self._max_jobs = {
            '_rfm_local': rt.runtime().get_option('systems/0/max_local_jobs')
        }
        self._pipeline_statistics = rt.runtime().get_option(
            'systems/0/dump_pipeline_progress'
        )
        self.task_listeners.append(self)

    def _init_pipeline_progress(self, num_tasks):
        self._pipeline_progress = {
            'startup': [(num_tasks, 0)],
            'ready_compile': [(0, 0)],
            'compiling': [(0, 0)],
            'ready_run': [(0, 0)],
            'running': [(0, 0)],
            'completing': [(0, 0)],
            'retired': [(0, 0)],
            'completed': [(0, 0)],
            'fail': [(0, 0)],
            'skip': [(0, 0)]
        }
        self._pipeline_step = 0
        self._t_pipeline_start = time.time()

    def _update_pipeline_progress(self, old_state, new_state, num_tasks=1):
        timestamp = time.time() - self._t_pipeline_start
        for state in self._pipeline_progress:
            count = self._pipeline_progress[state][self._pipeline_step][0]
            if old_state != new_state:
                if state == old_state:
                    count -= num_tasks
                elif state == new_state:
                    count += num_tasks

            self._pipeline_progress[state].append((count, timestamp))

        self._pipeline_step += 1

    def _dump_pipeline_progress(self, filename):
        import reframe.utility.jsonext as jsonext

        with open(filename, 'w') as fp:
            jsonext.dump(self._pipeline_progress, fp, indent=2)

    def runcase(self, case):
        super().runcase(case)
        check, partition, environ = case
        self._schedulers[partition.fullname] = partition.scheduler

        # Set partition-based counters, if not set already
        self._partition_tasks.setdefault(partition.fullname, util.OrderedSet())
        self._max_jobs.setdefault(partition.fullname, partition.max_jobs)

        task = RegressionTask(case, self.task_listeners)
        self._task_index[case] = task
        self.stats.add_task(task)
        getlogger().debug2(
            f'Added {check.name} on {partition.fullname} '
            f'using {environ.name}'
        )
        self._current_tasks.add(task)

    def exit(self):
        if self._pipeline_statistics:
            self._init_pipeline_progress(len(self._current_tasks))

        self._pollctl.reset_snooze_time()
        while self._current_tasks:
            try:
                self._poll_tasks()
                num_running = sum(
                    1 if t.state in ('running', 'compiling') else 0
                    for t in self._current_tasks
                )

                timeout = rt.runtime().get_option(
                    'general/0/pipeline_timeout'
                )

                self._advance_all(self._current_tasks, timeout)
                if self._pipeline_statistics:
                    num_retired = len(self._retired_tasks)

                _cleanup_all(self._retired_tasks, not self.keep_stage_files)
                if self._pipeline_statistics:
                    num_retired_actual = num_retired - len(self._retired_tasks)

                    # Some tests might not be cleaned up because they are
                    # waiting for dependencies or because their dependencies
                    # have failed.
                    self._update_pipeline_progress(
                        'retired', 'completed', num_retired_actual
                    )

                if num_running:
                    self._pollctl.snooze()
            except ABORT_REASONS as e:
                self._failall(e)
                raise

        if self._pipeline_statistics:
            self._dump_pipeline_progress('pipeline-progress.json')

    def _poll_tasks(self):
        for partname, sched in self._schedulers.items():
            jobs = []
            for t in self._partition_tasks[partname]:
                if t.state == 'compiling':
                    jobs.append(t.check.build_job)
                elif t.state == 'running':
                    jobs.append(t.check.job)

            sched.poll(*jobs)

    def _exec_stage(self, task, stage_methods):
        '''Execute a series of pipeline stages.

        Return True on success, False otherwise.
        '''

        try:
            for stage in stage_methods:
                stage()
        except TaskExit:
            self._current_tasks.remove(task)
            if task.check.current_partition:
                partname = task.check.current_partition.fullname
            else:
                partname = None

            # Remove tasks from the partition tasks if there
            with contextlib.suppress(KeyError):
                self._partition_tasks['_rfm_local'].remove(task)
                if partname:
                    self._partition_tasks[partname].remove(task)

            return False
        else:
            return True

    def _advance_all(self, tasks, timeout=None):
        t_init = time.time()
        num_progressed = 0

        getlogger().debug2(f'Current tests: {len(tasks)}')

        # We take a snapshot of the tasks to advance by doing a shallow copy,
        # since the tasks may removed by the individual advance functions.
        for t in list(tasks):
            old_state = t.state
            bump_state = getattr(self, f'_advance_{t.state}')
            num_progressed += bump_state(t)
            new_state = t.state

            t_elapsed = time.time() - t_init
            if timeout and t_elapsed > timeout and num_progressed:
                break

            if self._pipeline_statistics:
                self._update_pipeline_progress(old_state, new_state, 1)

        getlogger().debug2(f'Bumped {num_progressed} test(s)')

    def _advance_startup(self, task):
        if self.deps_skipped(task):
            try:
                raise SkipTestError('skipped due to skipped dependencies')
            except SkipTestError as e:
                task.skip()
                self._current_tasks.remove(task)
                return 1
        elif self.deps_succeeded(task):
            try:
                self.printer.status('RUN', task.info())
                task.setup(task.testcase.partition,
                           task.testcase.environ,
                           sched_flex_alloc_nodes=self.sched_flex_alloc_nodes,
                           sched_options=self.sched_options)
            except TaskExit:
                self._current_tasks.remove(task)
                return 1

            if isinstance(task.check, RunOnlyRegressionTest):
                # All tests should execute all the pipeline stages, even if
                # they are no-ops
                self._exec_stage(task, [task.compile,
                                        task.compile_complete,
                                        task.compile_wait])

            return 1
        elif self.deps_failed(task):
            exc = TaskDependencyError('dependencies failed')
            task.fail((type(exc), exc, None))
            self._current_tasks.remove(task)
            return 1
        else:
            # Not all dependencies have finished yet
            getlogger().debug2(f'{task.info()} waiting for dependencies')
            return 0

    def _advance_ready_compile(self, task):
        partname = _get_partition_name(task, phase='build')
        max_jobs = self._max_jobs[partname]
        if len(self._partition_tasks[partname]) < max_jobs:
            if self._exec_stage(task, [task.compile]):
                self._partition_tasks[partname].add(task)

            return 1

        getlogger().debug2(f'Hit the max job limit of {partname}: {max_jobs}')
        return 0

    def _advance_compiling(self, task):
        partname = _get_partition_name(task, phase='build')
        try:
            if task.compile_complete():
                task.compile_wait()
                self._partition_tasks[partname].remove(task)
                if isinstance(task.check, CompileOnlyRegressionTest):
                    # All tests should pass from all the pipeline stages,
                    # even if they are no-ops
                    self._exec_stage(task, [task.run,
                                            task.run_complete,
                                            task.run_wait])

                return 1
            else:
                return 0
        except TaskExit:
            self._partition_tasks[partname].remove(task)
            self._current_tasks.remove(task)
            return 1

    def _advance_ready_run(self, task):
        partname = _get_partition_name(task, phase='run')
        max_jobs = self._max_jobs[partname]
        if len(self._partition_tasks[partname]) < max_jobs:
            if self._exec_stage(task, [task.run]):
                self._partition_tasks[partname].add(task)

            return 1

        getlogger().debug2(f'Hit the max job limit of {partname}: {max_jobs}')
        return 0

    def _advance_running(self, task):
        partname = _get_partition_name(task, phase='run')
        try:
            if task.run_complete():
                if self._exec_stage(task, [task.run_wait]):
                    self._partition_tasks[partname].remove(task)

                return 1
            else:
                return 0
        except TaskExit:
            self._partition_tasks[partname].remove(task)
            self._current_tasks.remove(task)
            return 1

    def _advance_completing(self, task):
        try:
            if not self.skip_sanity_check:
                task.sanity()

            if not self.skip_performance_check:
                task.performance()

            task.finalize()
            self._retired_tasks.append(task)
            self._current_tasks.remove(task)
            return 1
        except TaskExit:
            self._current_tasks.remove(task)
            return 1

    def deps_failed(self, task):
        # NOTE: Restored dependencies are not in the task_index
        return any(self._task_index[c].failed
                   for c in task.testcase.deps if c in self._task_index)

    def deps_succeeded(self, task):
        # NOTE: Restored dependencies are not in the task_index
        return all(self._task_index[c].succeeded
                   for c in task.testcase.deps if c in self._task_index)

    def deps_skipped(self, task):
        # NOTE: Restored dependencies are not in the task_index
        return any(self._task_index[c].skipped
                   for c in task.testcase.deps if c in self._task_index)

    def _failall(self, cause):
        '''Mark all tests as failures'''
        getlogger().debug2(f'Aborting all tasks due to {type(cause).__name__}')
        for task in self._current_tasks:
            with contextlib.suppress(FailureLimitError):
                task.abort(cause)

    # These function can be useful for tracking statistics of the framework,
    # such as number of tests that have finished setup etc.
    def on_task_setup(self, task):
        pass

    def on_task_run(self, task):
        pass

    def on_task_compile(self, task):
        pass

    def on_task_exit(self, task):
        self._pollctl.reset_snooze_time()

    def on_task_compile_exit(self, task):
        self._pollctl.reset_snooze_time()

    def on_task_skip(self, task):
        msg = str(task.exc_info[1])
        self.printer.status('SKIP', msg, just='right')

    def on_task_failure(self, task):
        self._num_failed_tasks += 1
        msg = f'{task.info()}'
        if task.failed_stage == 'cleanup':
            self.printer.status('ERROR', msg, just='right')
        else:
            self.printer.status('FAIL', msg, just='right')

        _print_perf(task)
        timings = task.pipeline_timings(['setup',
                                         'compile_complete',
                                         'run_complete',
                                         'sanity',
                                         'performance',
                                         'total'])
        getlogger().info(f'==> test failed during {task.failed_stage!r}: '
                         f'test staged in {task.check.stagedir!r}')
        getlogger().verbose(f'==> {timings}')
        if self._num_failed_tasks >= self.max_failures:
            raise FailureLimitError(
                f'maximum number of failures ({self.max_failures}) reached'
            )

    def on_task_success(self, task):
        msg = f'{task.info()}'
        self.printer.status('OK', msg, just='right')
        _print_perf(task)
        timings = task.pipeline_timings(['setup',
                                         'compile_complete',
                                         'run_complete',
                                         'sanity',
                                         'performance',
                                         'total'])
        getlogger().verbose(f'==> {timings}')
        for c in task.testcase.deps:
            # NOTE: Restored dependencies are not in the task_index
            if c in self._task_index:
                self._task_index[c].ref_count -= 1



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

#
# Regression test loader
#

import ast
import contextlib
import inspect
import os
import sys
import traceback

import reframe.core.fields as fields
import reframe.utility as util
import reframe.utility.osext as osext
from reframe.core.exceptions import NameConflictError, is_severe, what
from reframe.core.logging import getlogger, time_function


class RegressionCheckValidator(ast.NodeVisitor):
    def __init__(self):
        self._has_import = False
        self._has_regression_test = False

    @property
    def valid(self):
        return self._has_import or self._has_regression_test

    def visit_Import(self, node):
        for m in node.names:
            if m.name.startswith('reframe'):
                self._has_import = True

    def visit_ImportFrom(self, node):
        if node.module is not None and node.module.startswith('reframe'):
            self._has_import = True

    def visit_ClassDef(self, node):
        for deco in node.decorator_list:
            with contextlib.suppress(AttributeError):
                if deco.attr == 'simple_test':
                    self._has_regression_test = True
                    break


class RegressionCheckLoader:
    def __init__(self, load_path, recurse=False, external_vars=None,
                 skip_system_check=False, skip_prgenv_check=False):
        # Expand any environment variables and symlinks
        load_path = [os.path.realpath(osext.expandvars(p)) for p in load_path]
        self._load_path = osext.unique_abs_paths(load_path, recurse)
        self._recurse = recurse

        # Loaded tests by name; maps test names to the file that were defined
        self._loaded = {}

        # Variables set in the command line
        self._external_vars = external_vars or {}
        self._unset_vars = {}
        self._skip_system_check = bool(skip_system_check)
        self._skip_prgenv_check = bool(skip_prgenv_check)

    def unset_vars(self, testname):
        return self._unset_vars.get(testname, [])

    def _module_name(self, filename):
        '''Figure out a module name from filename.

        If filename is an absolute path, module name will the basename without
        the extension. Otherwise, it will be the same as path with `/' replaced
        by `.' and without the final file extension.'''
        if os.path.isabs(filename):
            return os.path.splitext(os.path.basename(filename))[0]
        else:
            return (os.path.splitext(filename)[0]).replace('/', '.')

    def _validate_source(self, filename):
        '''Check if `filename` is a valid Reframe source file.'''

        with open(filename, 'r') as f:
            source_tree = ast.parse(f.read(), filename)

        msg = f'Validating {filename!r}: '
        validator = RegressionCheckValidator()
        validator.visit(source_tree)
        if validator.valid:
            msg += 'OK'
        else:
            msg += 'not a test file'

        getlogger().debug(msg)
        return validator.valid

    def _validate_check(self, check):
        import reframe.utility as util

        name = type(check).__name__
        checkfile = os.path.relpath(inspect.getfile(type(check)))
        required_attrs = ['valid_systems', 'valid_prog_environs']
        for attr in required_attrs:
            if not hasattr(check, attr):
                getlogger().warning(
                    f'skipping test {name!r}: {attr!r} not defined'
                )
                return False

        is_copyable = util.attr_validator(lambda obj: util.is_copyable(obj))
        valid, attr = is_copyable(check)
        if not valid:
            getlogger().warning(
                f'{checkfile}: {attr!r} is not copyable; '
                f'not copyable attributes are not '
                f'allowed inside the __init__() method or post-init hooks; '
                f'consider setting them in another pipeline hook instead'
            )
            return False

        return True

    @property
    def load_path(self):
        return self._load_path

    @property
    def prefix(self):
        return self._prefix

    @property
    def recurse(self):
        return self._recurse

    def _set_defaults(self, test_registry):
        if test_registry is None:
            return

        self._unset_vars = {}
        for test in test_registry:
            for name, val in self._external_vars.items():
                if '.' in name:
                    testname, varname = name.split('.', maxsplit=1)
                else:
                    testname, varname = test.__name__, name

                if testname == test.__name__:
                    # Treat special values
                    if val == '@none':
                        val = None
                    else:
                        val = fields.make_convertible(val)

                    if not test.setvar(varname, val):
                        self._unset_vars.setdefault(test.__name__, [])
                        self._unset_vars[test.__name__].append(varname)

    def load_from_module(self, module):
        '''Load user checks from module.

        This method tries to load the test registry from a given module and
        instantiates all the tests in the registry. The instantiated checks
        are validated before return.
        '''
        registry = getattr(module, '_rfm_test_registry', None)
        self._set_defaults(registry)
        reset_sysenv = self._skip_prgenv_check << 1 | self._skip_system_check
        if registry:
            candidate_tests = registry.instantiate_all(reset_sysenv)
        else:
            candidate_tests = []

        # Post-instantiation validation of the candidate tests
        final_tests = []
        for c in candidate_tests:
            if not self._validate_check(c):
                continue

            testfile = module.__file__
            try:
                conflicted = self._loaded[c.unique_name]
            except KeyError:
                self._loaded[c.unique_name] = testfile
                final_tests.append(c)
            else:
                raise NameConflictError(
                    f'test {c.unique_name!r} from {testfile!r} '
                    f'is already defined in {conflicted!r}'
                )

        getlogger().debug(f'  > Loaded {len(final_tests)} test(s)')
        return final_tests

    def load_from_file(self, filename, force=False):
        if not self._validate_source(filename):
            return []

        try:
            return self.load_from_module(
                util.import_module_from_file(filename, force)
            )
        except Exception:
            exc_info = sys.exc_info()
            if not is_severe(*exc_info):
                # Simply skip the file in this case
                getlogger().warning(
                    f"skipping test file {filename!r}: {what(*exc_info)} "
                    f"(rerun with '-v' for more information)"
                )
                getlogger().verbose(traceback.format_exc())
                return []
            else:
                raise

    def load_from_dir(self, dirname, recurse=False, force=False):
        checks = []
        for entry in os.scandir(dirname):
            if recurse and entry.is_dir():
                checks += self.load_from_dir(entry.path, recurse, force)

            if (entry.name.startswith('.') or
                not entry.name.endswith('.py') or
                not entry.is_file()):
                continue

            checks += self.load_from_file(entry.path, force)

        return checks

    @time_function
    def load_all(self, force=False):
        '''Load all checks in self._load_path.

        If a prefix exists, it will be prepended to each path.

        :arg force: Force reloading of test files.
        :returns: The list of loaded tests.
        '''
        checks = []
        for d in self._load_path:
            getlogger().debug(f'Looking for tests in {d!r}')
            if not os.path.exists(d):
                getlogger().warning(f'check path {d!r} does not exist')
                continue

            if os.path.isdir(d):
                checks += self.load_from_dir(d, self._recurse, force)
            else:
                checks += self.load_from_file(d, force)

        return checks



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

#
# Useful descriptors for advanced operations on fields
#

import datetime
import re

import reframe.utility.typecheck as types
from reframe.core.warnings import user_deprecation_warning
from reframe.utility import ScopedDict


class _Convertible:
    '''Wrapper for values that allowed to be converted implicitly'''

    __slots__ = ('value')

    def __init__(self, value):
        self.value = value


def make_convertible(value):
    return _Convertible(value)


def remove_convertible(value):
    if isinstance(value, _Convertible):
        return value.value
    else:
        return value


class Field:
    '''Base class for attribute validators.'''

    def __init__(self, attr_name=None):
        if attr_name is not None:
            self._name = attr_name

    def __set_name__(self, owner, name):
        if not hasattr(self, '_name'):
            self._name = name

    def __get__(self, obj, objtype):
        if obj is None:
            return self

        try:
            return obj.__dict__[self._name]
        except KeyError:
            # We raise an AttributeError to emulate the standard attribute
            # access.
            raise AttributeError("%s object has no attribute '%s'" %
                                 (objtype.__name__, self._name)) from None

    def __set__(self, obj, value):
        obj.__dict__[self._name] = remove_convertible(value)


class TypedField(Field):
    '''Stores a field of predefined type'''

    def __init__(self, main_type, *other_types, attr_name=None):
        super().__init__(attr_name)
        self._types = (main_type,) + other_types
        if not all(isinstance(t, type) for t in self._types):
            raise TypeError('{0} is not a sequence of types'.
                            format(self._types))

    @property
    def valid_types(self):
        return self._types

    def _check_type(self, value):
        if not any(isinstance(value, t) for t in self._types):
            typedescr = '|'.join(t.__name__ for t in self._types)
            raise TypeError(
                "failed to set field '%s': '%s' is not of type '%s'" %
                (self._name, value, typedescr))

    def __set__(self, obj, value):
        try:
            self._check_type(value)
        except TypeError:
            raw_value = remove_convertible(value)
            if raw_value is value:
                # value was not convertible; reraise
                raise

            # Try to convert value to any of the supported types
            value = raw_value
            for t in self._types:
                try:
                    value = t(value)
                except TypeError:
                    continue
                else:
                    super().__set__(obj, value)
                    return

            # Conversion failed
            typenames = [t.__name__ for t in self._types]
            raise TypeError(
                f'failed to set field {self._name!r}: '
                f'could not convert to any of the supported types: '
                f'{typenames}'
            )
        else:
            super().__set__(obj, value)


class ConstantField(Field):
    '''Holds a constant.

    Attempt to set it will raise an exception. This field may be accessed also
    from the class and will return the same constant value.

    :arg value: the value of this field.

    '''

    def __set_name__(self, owner, name):
        pass

    def __init__(self, value):
        self._value = value

    def __get__(self, obj, objtype):
        return self._value

    def __set__(self, obj, value):
        raise ValueError('attempt to set a read-only variable')


class TimerField(TypedField):
    '''Stores a timer in the form of a :class:`datetime.timedelta` object'''

    def __init__(self, *other_types, attr_name=None):
        super().__init__(str, int, float, *other_types, attr_name=attr_name)

    def __set__(self, obj, value):
        value = remove_convertible(value)
        self._check_type(value)
        if isinstance(value, str):
            time_match = re.match(r'^((?P<days>\d+)d)?'
                                  r'((?P<hours>\d+)h)?'
                                  r'((?P<minutes>\d+)m)?'
                                  r'((?P<seconds>\d+)s)?$',
                                  value)
            if not time_match:
                raise ValueError('invalid format for timer field')

            value = datetime.timedelta(
                **{k: int(v) for k, v in time_match.groupdict().items() if v}
            ).total_seconds()
        elif isinstance(value, float) or isinstance(value, int):
            if value < 0:
                raise ValueError('timer field value cannot be negative')

        # Call Field's __set__() method, type checking is already performed
        Field.__set__(self, obj, value)


class ScopedDictField(TypedField):
    '''Stores a ScopedDict with a specific type.

    It also handles implicit conversions from ordinary dicts.'''

    def __init__(self, valuetype, *other_types, attr_name=None):
        super().__init__(types.Dict[str, types.Dict[str, valuetype]],
                         ScopedDict, *other_types, attr_name=attr_name)

    def __set__(self, obj, value):
        value = remove_convertible(value)
        self._check_type(value)
        if not isinstance(value, ScopedDict):
            value = ScopedDict(value) if value is not None else value

        Field.__set__(self, obj, value)


class DeprecatedField(Field):
    '''Field wrapper for deprecating fields.'''

    OP_GET = 1
    OP_SET = 2
    OP_ALL = OP_SET | OP_GET

    @property
    def message(self):
        return self._message

    @property
    def op(self):
        return self._op

    @property
    def from_version(self):
        return self._from_version

    def __set_name__(self, owner, name):
        self._target_field.__set_name__(owner, name)

    def __init__(self, target_field, message, op=OP_ALL, from_version='0.0.0'):
        self._target_field = target_field
        self._message = message
        self._op = op
        self._from_version = from_version

    def __set__(self, obj, value):
        if self._op & DeprecatedField.OP_SET:
            user_deprecation_warning(self._message, self._from_version)

        self._target_field.__set__(obj, value)

    def __get__(self, obj, objtype):
        if self._op & DeprecatedField.OP_GET:
            user_deprecation_warning(self._message, self._from_version)

        return self._target_field.__get__(obj, objtype)



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import archspec.cpu
import contextlib
import glob
import os
import re

import reframe.utility.osext as osext
from reframe.core.exceptions import SpawnedProcessError


def _bits_from_str(mask_s):
    '''Return the set bits from a string representing a bit array.'''

    bits = []
    mask = int(mask_s, 0)
    pos = 0
    while mask:
        if mask & 1:
            bits.append(pos)

        pos += 1
        mask >>= 1

    return bits


def _str_from_bits(bits):
    '''Return a string representation of a bit array with ``bits`` set.'''

    ret = 0
    for b in bits:
        ret |= (1 << b)

    return hex(ret).lower()


def _sysfs_topo():
    cache_units = {
        'K': 1024,
        'M': 1024*1024,
        'G': 1024*1024*1024
    }
    cpuinfo = {
        'topology': {}
    }
    cpu_dirs = glob.glob(r'/sys/devices/system/cpu/cpu[0-9]*')
    nodes = glob.glob(r'/sys/devices/system/node/node[0-9]*')
    cores = set()
    for cpu in cpu_dirs:
        core_cpus_path = os.path.join(cpu, 'topology/core_cpus')
        thread_siblings_path = os.path.join(cpu, 'topology/thread_siblings')
        if glob.glob(core_cpus_path):
            cores_path = core_cpus_path
        elif glob.glob(thread_siblings_path):
            cores_path = thread_siblings_path
        else:
            # Information cannot be retrieved
            continue

        with contextlib.suppress(IOError):
            with open(cores_path) as fp:
                core_cpus = fp.read()
                core_cpus = re.sub(r'[\s,]', '', core_cpus)
                core_cpus = f'0x{core_cpus.lower()}'
                cores.add(core_cpus)

    sockets = set()
    for cpu in cpu_dirs:
        package_cpus_path = os.path.join(cpu, 'topology/package_cpus')
        core_siblings_path = os.path.join(cpu, 'topology/core_siblings')
        if glob.glob(package_cpus_path):
            sockets_path = package_cpus_path
        elif glob.glob(core_siblings_path):
            sockets_path = core_siblings_path
        else:
            # Information cannot be retrieved
            continue

        with contextlib.suppress(IOError):
            with open(sockets_path) as fp:
                package_cpus = fp.read()
                package_cpus = re.sub(r'[\s,]', '', package_cpus)
                package_cpus = f'0x{package_cpus.lower()}'
                sockets.add(package_cpus)

    numa_nodes = []
    for node in nodes:
        with contextlib.suppress(IOError):
            with open(os.path.join(node, 'cpumap')) as fp:
                cpumap = fp.read()
                cpumap = re.sub(r'[\s,]', '', cpumap)
                cpumap = f'0x{cpumap.lower()}'
                numa_nodes.append(cpumap)

    numa_nodes.sort()
    caches = {}
    for cpu in cpu_dirs:
        cache_dirs = glob.glob(cpu + r'/cache/index[0-9]*')
        for cache in cache_dirs:
            cache_level = 0
            cache_size = 0
            cache_linesize = 0
            cache_associativity = 0
            cache_cpuset = ''
            with contextlib.suppress(IOError):
                with open(os.path.join(cache, 'level')) as fp:
                    cache_level = int(fp.read())

            with contextlib.suppress(IOError):
                # Skip L1 instruction cache
                with open(os.path.join(cache, 'type')) as fp:
                    if cache_level == 1 and fp.read() == 'Instruction\n':
                        continue

            with contextlib.suppress(IOError):
                with open(os.path.join(cache, 'size')) as fp:
                    cache_size = fp.read()
                    m = re.match(r'(?P<val>\d+)(?P<unit>\S)', cache_size)
                    if m:
                        value = int(m.group('val'))
                        unit = cache_units.get(m.group('unit'), 1)
                        cache_size = value*unit

            with contextlib.suppress(IOError):
                with open(os.path.join(cache, 'coherency_line_size')) as fp:
                    cache_linesize = int(fp.read())

            # Don't take the associativity directly from
            # "ways_of_associativity" file because  some archs (ia64, ppc)
            # put 0 there when fully-associative, while others (x86)
            # put something like -1.
            with contextlib.suppress(IOError):
                with open(os.path.join(cache, 'number_of_sets')) as fp:
                    cache_number_of_sets = int(fp.read())

                with open(os.path.join(cache,
                                       'physical_line_partition')) as fp:
                    cache_physical_line_partition = int(fp.read())

                if (cache_linesize and
                    cache_physical_line_partition and
                    cache_number_of_sets):
                    cache_associativity = (cache_size //
                                           cache_linesize //
                                           cache_physical_line_partition //
                                           cache_number_of_sets)

            with contextlib.suppress(IOError):
                with open(os.path.join(cache, 'shared_cpu_map')) as fp:
                    cache_cpuset = fp.read()
                    cache_cpuset = re.sub(r'[\s,]', '', cache_cpuset)
                    cache_cpuset = f'0x{cache_cpuset.lower()}'

            num_cpus = len(_bits_from_str(cache_cpuset))
            caches.setdefault((cache_level, cache_size, cache_linesize,
                               cache_associativity, num_cpus), set())
            caches[(cache_level, cache_size, cache_linesize,
                    cache_associativity, num_cpus)].add(cache_cpuset)

    num_cpus = len(cpu_dirs)
    num_cores = len(cores)
    num_sockets = len(sockets)
    num_cpus_per_core = num_cpus // num_cores if num_cores else 0
    num_cpus_per_socket = num_cpus // num_sockets if num_sockets else 0

    # Fill in the cpuinfo
    cpuinfo['num_cpus'] = num_cpus
    cpuinfo['num_cpus_per_core'] = num_cpus_per_core
    cpuinfo['num_cpus_per_socket'] = num_cpus_per_socket
    cpuinfo['num_sockets'] = num_sockets
    cpuinfo['topology']['numa_nodes'] = numa_nodes
    cpuinfo['topology']['sockets'] = sorted(list(sockets))
    cpuinfo['topology']['cores'] = sorted(list(cores))
    cpuinfo['topology']['caches'] = []
    for cache_type, cpusets in caches.items():
        (cache_level, cache_size,
         cache_linesize, cache_associativity, num_cpus) = cache_type
        c = {
            'type': f'L{cache_level}',
            'size': cache_size,
            'linesize': cache_linesize,
            'associativity': cache_associativity,
            'num_cpus': num_cpus,
            'cpusets': sorted(list(cpusets))
        }
        cpuinfo['topology']['caches'].append(c)

    return cpuinfo


def _sysctl_topo():
    try:
        exec_output = osext.run_command('sysctl hw machdep.cpu',
                                        check=True)
    except (FileNotFoundError, SpawnedProcessError):
        return {}

    cpuinfo = {
        'topology': {}
    }
    match = re.search(r'hw\.ncpu: (?P<num_cpus>\d+)', exec_output.stdout)
    if match:
        num_cpus = int(match.group('num_cpus'))

    match = re.search(r'hw\.physicalcpu: (?P<num_cores>\d+)',
                      exec_output.stdout)
    if match:
        num_cores = int(match.group('num_cores'))

    match = re.search(r'hw\.packages: (?P<num_sockets>\d+)',
                      exec_output.stdout)
    if match:
        num_sockets = int(match.group('num_sockets'))
        cpuinfo['num_sockets'] = num_sockets

    match = re.search(r'hw\.cacheconfig:(?P<cacheconfig>(\s\d+)*)',
                      exec_output.stdout)
    if match:
        cacheconfig = list(map(int, match.group('cacheconfig').split()))

    match = re.search(r'hw\.cachesize:(?P<cachesize>(\s\d+)*)',
                      exec_output.stdout)
    if match:
        cachesize = list(map(int, match.group('cachesize').split()))

    match = re.search(r'hw\.cachelinesize: (?P<linesize>\d+)',
                      exec_output.stdout)
    if match:
        linesize = int(match.group('linesize'))

    # index 0 is referring to memory
    cache_associativity = [0]
    for i in range(1, len(cachesize)):
        if cachesize[i] == 0:
            break

        match = re.search(rf'machdep\.cpu\.cache\.L{i}_associativity: '
                          rf'(?P<associativity>\d+)',
                          exec_output.stdout)
        assoc = int(match.group('associativity')) if match else 0
        cache_associativity.append(assoc)

    num_cpus_per_socket = num_cpus // num_sockets
    num_cpus_per_core = num_cpus // num_cores

    # Fill in the cpuinfo
    cpuinfo['num_cpus'] = num_cpus
    cpuinfo['num_cpus_per_socket'] = num_cpus_per_socket
    cpuinfo['num_cpus_per_core'] = num_cpus_per_core
    cpuinfo['topology']['numa_nodes'] = [_str_from_bits(range(num_cpus))]
    cpuinfo['topology']['sockets'] = [
        _str_from_bits(range(start, start+num_cpus_per_socket))
        for start in range(0, num_cpus, num_cpus_per_socket)
    ]
    cpuinfo['topology']['cores'] = [
        _str_from_bits(range(start, start+num_cpus_per_core))
        for start in range(0, num_cpus, num_cpus_per_core)
    ]
    cpuinfo['topology']['caches'] = []
    for i in range(1, len(cache_associativity)):
        t = {
            'type': f'L{i}',
            'size': cachesize[i],
            'linesize': linesize,
            'associativity': cache_associativity[i],
            'num_cpus': cacheconfig[i],
            'cpusets': [
                _str_from_bits(range(start, start+cacheconfig[i]))
                for start in range(0, num_cpus, cacheconfig[i])
            ]
        }
        cpuinfo['topology']['caches'].append(t)

    return cpuinfo


def cpuinfo():
    ret = {
        'arch': archspec.cpu.host().name
    }

    # Try first to get information from the filesystem
    if os.path.isdir('/sys'):
        topology = _sysfs_topo()
    else:
        # Try with the `sysctl` command
        topology = _sysctl_topo()

    ret.update(topology)
    return ret



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause


import reframe.core.builtins as builtins
import reframe.core.runtime as runtime
import reframe.utility as util

from reframe.core.decorators import TestRegistry
from reframe.core.logging import getlogger, time_function
from reframe.core.meta import make_test
from reframe.core.schedulers import Job
from reframe.frontend.executors import generate_testcases


@time_function
def getallnodes(state='all', jobs_cli_options=None):
    rt = runtime.runtime()
    nodes = {}
    for part in rt.system.partitions:
        # This job will not be submitted, it's used only to filter
        # the nodes based on the partition configuration
        dummy_job = Job.create(part.scheduler,
                               part.launcher_type(),
                               name='placeholder-job',
                               sched_access=part.access,
                               sched_options=jobs_cli_options)

        available_nodes = part.scheduler.allnodes()
        available_nodes = part.scheduler.filternodes(dummy_job,
                                                     available_nodes)
        getlogger().debug(
            f'Total available nodes for {part.name}: {len(available_nodes)}'
        )

        if state.casefold() != 'all':
            available_nodes = {n for n in available_nodes
                               if n.in_state(state)}
            getlogger().debug(
                f'[F] Selecting nodes in state {state!r}: '
                f'available nodes now: {len(available_nodes)}'
            )

        nodes[part.fullname] = [n.name for n in available_nodes]

    return nodes


def _rfm_pin_run_nodes(obj):
    nodelist = getattr(obj, '$nid')
    if not obj.local:
        obj.job.pin_nodes = nodelist


def _rfm_pin_build_nodes(obj):
    pin_nodes = getattr(obj, '$nid')
    if not obj.local and not obj.build_locally:
        obj.build_job.pin_nodes = pin_nodes


def make_valid_systems_hook(systems):
    '''Returns a function to be used as a hook that sets the
    valid systems.

    Since valid_systems change for each generated test, we need to
    generate different post-init hooks for each one of them.
    '''
    def _rfm_set_valid_systems(obj):
        obj.valid_systems = systems

    return _rfm_set_valid_systems


@time_function
def distribute_tests(testcases, node_map):
    '''Returns new testcases that will be parameterized to run in node of
    their partitions based on the nodemap
    '''
    tmp_registry = TestRegistry()

    # We don't want to register the same check for every environment
    # per partition
    check_part_combs = set()
    for tc in testcases:
        check, partition, _ = tc
        candidate_comb = (check.unique_name, partition.fullname)
        if check.is_fixture() or candidate_comb in check_part_combs:
            continue

        check_part_combs.add(candidate_comb)
        cls = type(check)
        variant_info = cls.get_variant_info(
            check.variant_num, recurse=True
        )
        nc = make_test(
            f'{cls.__name__}_{partition.fullname.replace(":", "_")}',
            (cls,),
            {
                'valid_systems': [partition.fullname],
                '$nid': builtins.parameter(
                    [[n] for n in node_map[partition.fullname]],
                    fmt=util.nodelist_abbrev
                )
            },
            methods=[
                builtins.run_before('run')(_rfm_pin_run_nodes),
                builtins.run_before('compile')(_rfm_pin_build_nodes),
                # We re-set the valid system in a hook to make sure that it
                # will not be overwriten by a parent post-init hook
                builtins.run_after('init')(
                    make_valid_systems_hook([partition.fullname])
                ),
            ]
        )

        # We have to set the prefix manually
        nc._rfm_custom_prefix = check.prefix
        for i in range(nc.num_variants):
            # Check if this variant should be instantiated
            vinfo = nc.get_variant_info(i, recurse=True)
            vinfo['params'].pop('$nid')
            if vinfo == variant_info:
                tmp_registry.add(nc, variant_num=i)

    new_checks = tmp_registry.instantiate_all()
    return generate_testcases(new_checks)


@time_function
def repeat_tests(testcases, num_repeats):
    '''Returns new test cases parameterized over their repetition number'''

    tmp_registry = TestRegistry()
    unique_checks = set()
    for tc in testcases:
        check = tc.check
        if check.is_fixture() or check in unique_checks:
            continue

        unique_checks.add(check)
        cls = type(check)
        variant_info = cls.get_variant_info(
            check.variant_num, recurse=True
        )
        nc = make_test(
            f'{cls.__name__}', (cls,),
            {
                '$repeat_no': builtins.parameter(range(num_repeats))
            }
        )
        nc._rfm_custom_prefix = check.prefix
        for i in range(nc.num_variants):
            # Check if this variant should be instantiated
            vinfo = nc.get_variant_info(i, recurse=True)
            vinfo['params'].pop('$repeat_no')
            if vinfo == variant_info:
                tmp_registry.add(nc, variant_num=i)

    new_checks = tmp_registry.instantiate_all()
    return generate_testcases(new_checks)



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

'''Dynamic recursive type checking of collections.

This module defines types for collections, such as lists, dictionaries etc.,
that you can use with the :py:func:`isinstance` builtin function to
recursively type check all the elements of the collection. Suppose you have a
list of integers, suchs as ``[1, 2, 3]``, the following checks should be true:

.. code-block:: python

    l = [1, 2, 3]
    assert isinstance(l, List[int]) == True
    assert isinstance(l, List[float]) == False


Aggregate types can be combined in an arbitrary depth, so that you can type
check any complex data strcture:

.. code-block:: python

    d = {'a': [1, 2], 'b': [3, 4]}
    assert isisntance(d, Dict) == True
    assert isisntance(d, Dict[str, List[int]]) == True


This module offers the following aggregate types:

.. py:data:: List[T]

   A list with elements of type :class:`T`.

.. py:data:: Set[T]

   A set with elements of type :class:`T`.

.. py:data:: Dict[K,V]

   A dictionary with keys of type :class:`K` and values of type :class:`V`.

.. py:data:: Tuple[T]

   A tuple with elements of type :class:`T`.

.. py:data:: Tuple[T1,T2,...,Tn]

   A tuple with ``n`` elements, whose types are exactly :class:`T1`,
   :class:`T2`, ..., :class:`Tn` in that order.


.. py:data:: Str[patt]

   A string type whose members are all the strings matching the regular
   expression ``patt``.


Implementation details
----------------------

Internally, this module leverages metaclasses and the
:py:func:`__isinstancecheck__` method to customize the behaviour of the
:py:func:`isinstance` builtin.

By implementing also the :py:func:`__getitem__` accessor method, this module
follows the look-and-feel of the type hints proposed in `PEP484
<https://www.python.org/dev/peps/pep-0484/>`__. This method returns a new type
that is a subtype of the base container type. Using the facilities of
:py:class:`abc.ABCMeta`, builtin types, such as :py:class:`list`,
:py:class:`str` etc. are registered as subtypes of the base container types
offered by this module. The type hierarchy of the types defined in this module
is the following (example shown for :class:`List`, but it is analogous for
the rest of the types):

.. code-block:: none

          type
            |
            |
            |
          List
        /   |
       /    |
      /     |
    list  List[T]


In the above example :class:`T` may refer to any type, so that
:class:`List[List[int]]` is an instance of :class:`List`, but not an instance
of :class:`List[int]`.

'''

import abc
import re


class ConvertibleType(abc.ABCMeta):
    '''A type that support conversions from other types.

    This is a metaclass that allows classes that use it to support arbitrary
    conversions from other types using a cast-like syntax without having to
    change their constructor:

    .. code-block:: python

       new_obj = convertible_type(another_type)

    For example, a class whose constructor accepts and :class:`int` may need
    to support a cast-from-string conversion. This is particular useful if you
    want a custom-typed test
    :attr:`~reframe.core.pipeline.RegressionMixin.variable` to be able to be
    set from the command line using the :option:`-S` option.

    In order to support such conversions, a class must use this metaclass and
    define a class method, named as :obj:`__rfm_cast_<type>__`, for each of
    the type conversion that needs to support .

    The following is an example of a class :class:`X` that its normal
    constructor accepts two arguments but it also allows conversions from
    string:

    .. code-block:: python

       class X(metaclass=ConvertibleType):
           def __init__(self, x, y):
               self.data = (x, y)

           @classmethod
           def __rfm_cast_str__(cls, s):
               return X(*(int(x) for x in s.split(',', maxsplit=1)))

        assert X(2, 3).data == X('2,3').data

    .. versionadded:: 3.8.0

    '''

    def __call__(cls, *args, **kwargs):
        if len(args) == 1:
            cast_fn_name = f'__rfm_cast_{type(args[0]).__name__}__'
            if hasattr(cls, cast_fn_name):
                cast_fn = getattr(cls, cast_fn_name)
                return cast_fn(args[0])

        return super().__call__(*args, **kwargs)


# Metaclasses that implement the isinstance logic for the different builtin
# container types


class _CompositeType(abc.ABCMeta):
    def __instancecheck__(cls, inst):
        assert hasattr(cls, '_types') and len(cls._types) == 2
        return (issubclass(type(inst), cls._types[0]) or
                issubclass(type(inst), cls._types[1]))


class _InvertedType(abc.ABCMeta):
    def __instancecheck__(cls, inst):
        assert hasattr(cls, '_xtype')
        return not issubclass(type(inst), cls._xtype)


class _BuiltinType(ConvertibleType):
    def __init__(cls, name, bases, namespace):
        # Make sure that the class defines `_type`
        cls._bases = bases
        cls._namespace = namespace
        assert hasattr(cls, '_type')
        cls.register(cls._type)

    def __instancecheck__(cls, inst):
        if hasattr(cls, '_types'):
            return (issubclass(type(inst), cls._types[0]) or
                    issubclass(type(inst), cls._types[1]))

        if hasattr(cls, '_xtype'):
            return not issubclass(type(inst), cls._xtype)

        return issubclass(type(inst), cls)

    def __or__(cls, other):
        new_type = _BuiltinType(f'{cls.__name__}|{other.__name__}',
                                cls._bases, cls._namespace)
        new_type._types = (cls, other)
        return new_type

    def __invert__(cls):
        new_type = _BuiltinType(f'~{cls.__name__}',
                                cls._bases, cls._namespace)
        new_type._xtype = cls
        return new_type


class _SequenceType(_BuiltinType):
    '''A metaclass for containers with uniformly typed elements.'''

    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        cls._elem_type = None

    def __instancecheck__(cls, inst):
        if not super().__instancecheck__(inst):
            return False

        if cls._elem_type is None:
            return True

        return all(isinstance(c, cls._elem_type) for c in inst)

    def __getitem__(cls, elem_type):
        if not isinstance(elem_type, type):
            raise TypeError('{0} is not a valid type'.format(elem_type))

        if isinstance(elem_type, tuple):
            raise TypeError('invalid type specification for container type: '
                            'expected ContainerType[elem_type]')

        ret = _SequenceType('%s[%s]' % (cls.__name__, elem_type.__name__),
                            cls._bases, cls._namespace)
        ret._elem_type = elem_type
        cls.register(ret)
        return ret

    def __rfm_cast_str__(cls, s):
        container_type = cls._type
        elem_type = cls._elem_type
        return container_type(elem_type(e) for e in s.split(','))


class _TupleType(_SequenceType):
    '''A metaclass for tuples.

    Tuples may contain uniformly-typed elements or non-uniformly typed ones.
    '''

    def __instancecheck__(cls, inst):
        if not issubclass(type(inst), cls):
            return False

        if cls._elem_type is None:
            return True

        if len(cls._elem_type) == 1:
            # tuple with elements of the same type
            return all(isinstance(c, cls._elem_type[0]) for c in inst)

        # Non-uniformly typed tuple
        if len(inst) != len(cls._elem_type):
            return False

        return all(isinstance(elem, req_type)
                   for req_type, elem in zip(cls._elem_type, inst))

    def __getitem__(cls, elem_types):
        if not isinstance(elem_types, tuple):
            elem_types = (elem_types,)

        for t in elem_types:
            if not isinstance(t, type):
                raise TypeError('{0} is not a valid type'.format(t))

        cls_name = '%s[%s]' % (
            cls.__name__, ','.join(c.__name__ for c in elem_types)
        )
        ret = _TupleType(cls_name, cls._bases, cls._namespace)
        ret._elem_type = elem_types
        cls.register(ret)
        return ret

    def __rfm_cast_str__(cls, s):
        container_type = cls._type
        elem_types = cls._elem_type
        elems = s.split(',')
        if len(elem_types) == 1:
            elem_t = elem_types[0]
            return container_type(elem_t(e) for e in elems)
        elif len(elem_types) != len(elems):
            raise TypeError(f'cannot convert string {s!r} to {cls.__name__!r}')
        else:
            return container_type(
                elem_t(e) for elem_t, e in zip(elem_types, elems)
            )


class _MappingType(_BuiltinType):
    '''A metaclass for type checking mapping types.'''

    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        cls._key_type = None
        cls._value_type = None
        cls._bases = bases
        cls._namespace = namespace

    def __instancecheck__(cls, inst):
        if not issubclass(type(inst), cls):
            return False

        if cls._key_type is None and cls._key_type is None:
            return True

        assert cls._key_type is not None and cls._value_type is not None
        has_valid_keys = all(isinstance(k, cls._key_type)
                             for k in inst.keys())
        has_valid_values = all(isinstance(v, cls._value_type)
                               for v in inst.values())
        return has_valid_keys and has_valid_values

    def __getitem__(cls, typespec):
        try:
            key_type, value_type = typespec
        except ValueError:
            raise TypeError(
                'invalid type specification for mapping type: '
                'expected MappingType[key_type, value_type]') from None

        for t in typespec:
            if not isinstance(t, type):
                raise TypeError('{0} is not a valid type'.format(t))

        cls_name = '%s[%s,%s]' % (cls.__name__, key_type.__name__,
                                  value_type.__name__)
        ret = _MappingType(cls_name, cls._bases, cls._namespace)
        ret._key_type = key_type
        ret._value_type = value_type
        cls.register(ret)
        return ret

    def __rfm_cast_str__(cls, s):
        mappping_type = cls._type
        key_type = cls._key_type
        value_type = cls._value_type
        seq = []
        for key_datum in s.split(','):
            try:
                k, v = key_datum.split(':')
            except ValueError:
                # Re-raise as TypeError
                raise TypeError(
                    f'cannot convert string {s!r} to {cls.__name__!r}'
                ) from None

            seq.append((key_type(k), value_type(v)))

        return mappping_type(seq)


class _StrType(_SequenceType):
    '''A metaclass for type checking string types.'''

    def __instancecheck__(cls, inst):
        if not issubclass(type(inst), cls):
            return False

        if cls._elem_type is None:
            return True

        # _elem_type is a regex
        return re.fullmatch(cls._elem_type, inst) is not None

    def __getitem__(cls, patt):
        if not isinstance(patt, str):
            raise TypeError('invalid type specification for string type: '
                            'expected _StrType[regex]')

        ret = _StrType("%s[r'%s']" % (cls.__name__, patt),
                       cls._bases, cls._namespace)
        ret._elem_type = patt
        cls.register(ret)
        return ret

    def __rfm_cast_str__(cls, s):
        if not isinstance(s, cls):
            raise TypeError(f'cannot convert string {s!r} to {cls.__name__!r}')

        return s


class Bool(metaclass=_BuiltinType):
    '''A boolean type accepting implicit conversions from strings.

    This type represents a boolean value but allows implicit conversions from
    :class:`str`. More specifically, the following conversions are supported:

    - The strings ``'yes'``, ``'true'`` and ``'1'`` are converted to ``True``.
    - The strings ``'no'``, ``'false'`` and ``'0'`` are converted to
      ``False``.

    The built-in :class:`bool` type is registered as a subclass of this type.

    Boolean test variables that are meant to be set properly from the command
    line must be declared of this type and not :class:`bool`.

    '''

    _type = bool

    @classmethod
    def __rfm_cast_str__(cls, s):
        if s in ('true', 'yes', '1'):
            return True
        elif s in ('false', 'no', '0'):
            return False

        raise TypeError(f'cannot convert {s!r} to bool')


def make_meta_type(name, cls, metacls=_BuiltinType):
    namespace = metacls.__prepare__(name, ())
    namespace['_type'] = cls
    ret = metacls(name, (), namespace)
    return ret


Dict    = make_meta_type('Dict', dict, _MappingType)
Float   = make_meta_type('Float', float)
Integer = make_meta_type('Integer', int)
List    = make_meta_type('List', list, _SequenceType)
Set     = make_meta_type('Set', set, _SequenceType)
Str     = make_meta_type('Str', str, _StrType)
Tuple   = make_meta_type('Tuple', tuple, _TupleType)
